from typing import Optional, Union, cast

from pyspark.ml.evaluation import (
    BinaryClassificationEvaluator,
    RegressionEvaluator,
    MulticlassClassificationEvaluator,
    Evaluator,
)
from pyspark.ml.functions import vector_to_array
from pyspark.sql.pandas.functions import pandas_udf

from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.utils import SparkDataFrame
from sparklightautoml.tasks.losses.base import SparkLoss
from lightautoml.tasks import Task as LAMATask
from lightautoml.tasks.base import LAMLMetric, _default_losses

import pandas as pd
import numpy as np

DEFAULT_PREDICTION_COL_NAME = "prediction"
DEFAULT_TARGET_COL_NAME = "target"
DEFAULT_PROBABILITY_COL_NAME = "probability"


def argmax_in_vector(vec: pd.Series) -> pd.Series:
    vec.transform()


class SparkMetric(LAMLMetric):
    """
    Spark version of metric function that implements function assessing prediction error.
    """

    def __init__(
        self,
        name: str,
        metric_name: str,
        target_col: Optional[str] = None,
        prediction_col: Optional[str] = None,
        greater_is_better: bool = True,
    ):
        """

        Args:
            metric: Specifies metric. Format:
                ``func(y_true, y_false, Optional[sample_weight], **kwargs)`` -> `float`.
            name: Name of metric.
            greater_is_better: Whether or not higher metric value is better.
            one_dim: `True` for single class, False for multiclass.
            weighted: Weights of classes.
            **kwargs: Other parameters for metric.

        """
        self._name = name
        self._metric_name = metric_name
        self._target_col = target_col
        self._prediction_col = prediction_col
        self.greater_is_better = greater_is_better

    def __call__(self, dataset: Union[SparkDataset, SparkDataFrame], dropna: bool = False):
        sdf: SparkDataFrame = dataset.data if isinstance(dataset, SparkDataset) else dataset

        if self._target_col is not None and self._prediction_col is not None:
            assert self._target_col in sdf.columns and self._prediction_col in sdf.columns
            prediction_column = self._prediction_col
            target_column = self._target_col
        elif isinstance(dataset, SparkDataset):
            if len(dataset.features) == 1:
                prediction_column = dataset.features[0]
            else:
                prediction_column = next(c for c in dataset.data.columns if c.startswith("prediction"))
            target_column = dataset.target_column
        else:
            sdf = cast(SparkDataFrame, dataset)
            assert DEFAULT_PREDICTION_COL_NAME in sdf.columns and DEFAULT_TARGET_COL_NAME in sdf.columns
            prediction_column = DEFAULT_PREDICTION_COL_NAME
            target_column = DEFAULT_TARGET_COL_NAME

        sdf = sdf.dropna(subset=[prediction_column, target_column]) if dropna else sdf

        if self._name == "binary":
            evaluator = BinaryClassificationEvaluator(rawPredictionCol=prediction_column)
        elif self._name == "reg":
            evaluator = RegressionEvaluator(predictionCol=prediction_column)
        else:
            temp_pred_col = "multiclass_temp_prediction"
            evaluator = MulticlassClassificationEvaluator(probabilityCol=prediction_column, predictionCol=temp_pred_col)

            @pandas_udf("double")
            def argmax(vec: pd.Series) -> pd.Series:
                return vec.transform(lambda x: np.argmax(x))

            sdf = sdf.select("*", argmax(vector_to_array(prediction_column)).alias(temp_pred_col))

        evaluator = evaluator.setMetricName(self._metric_name).setLabelCol(target_column)

        score = evaluator.evaluate(sdf)
        sign = 2 * float(self.greater_is_better) - 1
        return score * sign

    @property
    def name(self) -> str:
        return self._name


class SparkTask(LAMATask):
    """
    Specify task (binary classification, multiclass classification, regression), metrics, losses.
    """

    _default_metrics = {"binary": "auc", "reg": "mse", "multiclass": "crossentropy"}

    _supported_metrics = {
        "binary": {"auc": "areaUnderROC", "aupr": "areaUnderPR"},
        "reg": {
            "r2": "rmse",
            "mse": "mse",
            "mae": "mae",
        },
        "multiclass": {
            "crossentropy": "logLoss",
            "accuracy": "accuracy",
            "f1_micro": "f1",
            "f1_weighted": "weightedFMeasure",
        },
    }

    def __init__(
        self,
        name: str,
        loss: Optional[str] = None,
        metric: Optional[str] = None,
        greater_is_better: Optional[bool] = None,
    ):

        super().__init__(name, loss, None, metric, None, greater_is_better)

        if metric is None:
            metric = self._default_metrics[name]

        # add losses
        # if None - infer from task
        self.losses = {}
        if loss is None:
            loss = _default_losses[self.name]
        # SparkLoss actualy does nothing, but it is there
        # to male TabularAutoML work
        self.losses = {"lgb": SparkLoss(), "linear_l2": SparkLoss()}

        assert metric in self._supported_metrics[self.name], (
            f"Unsupported metric {metric} for task {self.name}."
            f"The following metrics are supported: {list(self._supported_metrics[self.name].keys())}"
        )

        self.metric_name = metric

    def get_dataset_metric(self) -> LAMLMetric:
        """Obtains a function to calculate the metric on a dataset."""
        spark_metric_name = self._supported_metrics[self.name][self.metric_name]
        return SparkMetric(self.name, metric_name=spark_metric_name, greater_is_better=self.greater_is_better)



import functools
from abc import ABC
from copy import copy
from typing import Tuple, cast, Optional, List, Sequence

from pyspark.sql import functions as F

from lightautoml.pipelines.features.base import FeaturesPipeline
from lightautoml.pipelines.selection.base import SelectionPipeline
from lightautoml.reader.base import RolesDict
from sparklightautoml import VALIDATION_COLUMN
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.utils import SparkDataFrame
from sparklightautoml.pipelines.base import InputFeaturesAndRoles
from sparklightautoml.pipelines.features.base import SparkFeaturesPipeline
from lightautoml.validation.base import TrainValidIterator


class SparkBaseTrainValidIterator(TrainValidIterator, InputFeaturesAndRoles, ABC):
    """
    Implements applying selection pipeline and feature pipeline to SparkDataset.
    """

    TRAIN_VAL_COLUMN = VALIDATION_COLUMN

    def __init__(self, train: SparkDataset, input_roles: Optional[RolesDict] = None):
        assert train.folds_column in train.data.columns
        super().__init__(train)
        if not input_roles:
            input_roles = train.roles
        self._input_roles = input_roles

    def __next__(self) -> Tuple[SparkDataset, SparkDataset, SparkDataset]:
        """Define how to get next object.

        Returns:
            a tuple with:
            - full dataset (both train and valid parts with column containing the bool feature),
            - train part of the dataset
            - validation part of the dataset.

        """
        ...

    @property
    def features(self) -> List[str]:
        return self.input_features

    def apply_selector(self, selector: SelectionPipeline) -> "SparkBaseTrainValidIterator":
        """Select features on train data.

        Check if selector is fitted.
        If not - fit and then perform selection.
        If fitted, check if it's ok to apply.

        Args:
            selector: Uses for feature selection.

        Returns:
            Dataset with selected features.

        """
        sel_train_valid = copy(self)
        sel_train_valid.train = self.train[:, list(self.input_roles.keys())]

        if not selector.is_fitted:
            selector.fit(sel_train_valid)
            sfp = cast(SparkFeaturesPipeline, selector.features_pipeline)
            if sfp is not None:
                sfp.release_cache()

        train_valid = copy(self)
        # we don't need to create transformer for subselecting
        # because train_valid.input_roles is used in fit_... methods
        # of features pipelines and ml_algo to define columns they work with
        selected_features = selector.select(train_valid.train).features
        train_valid.input_roles = {
            feat: self.input_roles[feat]
            for feat in selected_features if feat in self.input_roles
        }

        return train_valid

    def apply_feature_pipeline(self, features_pipeline: SparkFeaturesPipeline) -> "SparkBaseTrainValidIterator":
        features_pipeline.input_roles = self.input_roles
        train_valid = cast(SparkBaseTrainValidIterator, super().apply_feature_pipeline(features_pipeline))
        train_valid.input_roles = features_pipeline.output_roles
        return train_valid

    def combine_val_preds(self, val_preds: Sequence[SparkDataFrame], include_train: bool = False) -> SparkDataFrame:
        # depending on train_valid logic there may be several ways of treating predictions results:
        # 1. for folds iterators - just union the results, it will yield the full train dataset
        # 2. for holdout iterators - create None predictions in train_part and union with valid part
        # 3. for custom iterators which may put the same records in
        #   different folds: union + groupby + (optionally) union with None-fied train_part
        # 4. for dummy - do nothing
        raise NotImplementedError()

    def _split_by_fold(self, fold: int) -> Tuple[SparkDataset, SparkDataset, SparkDataset]:
        train = cast(SparkDataset, self.train)
        is_val_col = (
            F.when(F.col(self.train.folds_column) != fold, F.lit(0)).otherwise(F.lit(1)).alias(self.TRAIN_VAL_COLUMN)
        )

        sdf = train.data.select("*", is_val_col)
        train_part_sdf = sdf.where(F.col(self.TRAIN_VAL_COLUMN) == 0).drop(self.TRAIN_VAL_COLUMN)
        valid_part_sdf = sdf.where(F.col(self.TRAIN_VAL_COLUMN) == 1).drop(self.TRAIN_VAL_COLUMN)

        train_ds = cast(SparkDataset, self.train.empty())
        train_ds.set_data(sdf, self.train.features, self.train.roles)

        train_part_ds = cast(SparkDataset, self.train.empty())
        train_part_ds.set_data(train_part_sdf, self.train.features, self.train.roles)

        valid_part_ds = cast(SparkDataset, self.train.empty())
        valid_part_ds.set_data(valid_part_sdf, self.train.features, self.train.roles)

        return train_ds, train_part_ds, valid_part_ds



from typing import List

from lightautoml.dataset.base import RolesDict


class InputFeaturesAndRoles:
    """
    Class that represents input features and input roles.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._input_roles = None

    @property
    def input_features(self) -> List[str]:
        """Returns list of input features"""
        return list(self.input_roles.keys())

    @property
    def input_roles(self) -> RolesDict:
        """Returns dict of input roles"""
        return self._input_roles

    @input_roles.setter
    def input_roles(self, roles: RolesDict):
        """Sets dict of input roles"""
        assert roles is None or isinstance(roles, dict)
        self._input_roles = roles


class OutputFeaturesAndRoles:
    """
    Class that represents output features and output roles.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._output_roles = None

    @property
    def output_features(self) -> List[str]:
        """Returns list of output features"""
        return list(self.output_roles.keys())

    @property
    def output_roles(self) -> RolesDict:
        """Returns dcit of output roles"""
        return self._output_roles



import functools
import logging
from typing import Optional, cast, Tuple, Iterable, Sequence

from lightautoml.dataset.base import LAMLDataset, RolesDict
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.utils import SparkDataFrame
from sparklightautoml.transformers.scala_wrappers.balanced_union_partitions_coalescer import (
    BalancedUnionPartitionsCoalescerTransformer,
)
from sparklightautoml.validation.base import SparkBaseTrainValidIterator
from lightautoml.validation.base import TrainValidIterator, HoldoutIterator

from pyspark.sql import functions as F


logger = logging.getLogger(__name__)


class SparkDummyIterator(SparkBaseTrainValidIterator):
    """
    Simple one step iterator over train part of SparkDataset
    """

    def __init__(self, train: SparkDataset, input_roles: Optional[RolesDict] = None):
        super().__init__(train, input_roles)
        self._curr_idx = 0

    def __iter__(self) -> Iterable:
        self._curr_idx = 0
        return self

    def __len__(self) -> Optional[int]:
        return 1

    def __next__(self) -> Tuple[SparkDataset, SparkDataset, SparkDataset]:
        """Define how to get next object.

        Returns:
            None, train dataset, validation dataset.

        """
        if self._curr_idx > 0:
            raise StopIteration

        self._curr_idx += 1

        sdf = cast(SparkDataFrame, self.train.data)
        sdf = sdf.withColumn(self.TRAIN_VAL_COLUMN, F.lit(0))

        train_ds = cast(SparkDataset, self.train.empty())
        train_ds.set_data(sdf, self.train.features, self.train.roles)

        return train_ds, train_ds, train_ds

    def combine_val_preds(self, val_preds: Sequence[SparkDataFrame], include_train: bool = False) -> SparkDataFrame:
        assert len(val_preds) == 1
        return val_preds[0]

    def get_validation_data(self) -> SparkDataset:
        return self.train

    def convert_to_holdout_iterator(self) -> "SparkHoldoutIterator":
        sds = cast(SparkDataset, self.train)
        assert sds.folds_column is not None, "Cannot convert to Holdout iterator when folds_column is not defined"
        return SparkHoldoutIterator(self.train, self.input_roles)


class SparkHoldoutIterator(SparkBaseTrainValidIterator):
    """Simple one step iterator over one fold of SparkDataset"""

    def __init__(self, train: SparkDataset, input_roles: Optional[RolesDict] = None):
        super().__init__(train, input_roles)
        self._curr_idx = 0

    def __iter__(self) -> Iterable:
        self._curr_idx = 0
        return self

    def __len__(self) -> Optional[int]:
        return 1

    def __next__(self) -> Tuple[SparkDataset, SparkDataset, SparkDataset]:
        """Define how to get next object.

        Returns:
            None, train dataset, validation dataset.

        """
        if self._curr_idx > 0:
            raise StopIteration

        full_ds, train_part_ds, valid_part_ds = self._split_by_fold(self._curr_idx)
        self._curr_idx += 1

        return full_ds, train_part_ds, valid_part_ds

    def get_validation_data(self) -> SparkDataset:
        full_ds, train_part_ds, valid_part_ds = self._split_by_fold(fold=0)
        return valid_part_ds

    def convert_to_holdout_iterator(self) -> "SparkHoldoutIterator":
        return self

    def combine_val_preds(self, val_preds: Sequence[SparkDataFrame], include_train: bool = False) -> SparkDataFrame:
        if len(val_preds) != 1:
            k = 0
        assert len(val_preds) == 1

        if not include_train:
            return val_preds[0]

        val_pred_cols = set(val_preds[0].columns)
        train_cols = set(self.train.columns)
        assert len(train_cols.difference(val_pred_cols)) == 0
        new_feats = val_pred_cols.difference(train_cols)

        _, train_ds, _ = self._split_by_fold(0)
        missing_cols = [F.lit(None).alias(f) for f in new_feats]
        full_val_preds = train_ds.select("*", *missing_cols).unionByName(val_preds[0])

        return full_val_preds


class SparkFoldsIterator(SparkBaseTrainValidIterator):
    """Classic cv iterator.

    Folds should be defined in Reader, based on cross validation method.
    """

    def __init__(self, train: SparkDataset, n_folds: Optional[int] = None, input_roles: Optional[RolesDict] = None):
        """Creates iterator.

        Args:
            train: Dataset for folding.
            n_folds: Number of folds.

        """
        super().__init__(train, input_roles)

        num_folds = train.data.select(F.max(train.folds_column).alias("max")).first()["max"]
        self.n_folds = num_folds + 1
        if n_folds is not None:
            self.n_folds = min(self.n_folds, n_folds)

    def __len__(self) -> int:
        """Get len of iterator.

        Returns:
            Number of folds.

        """
        return self.n_folds

    def __iter__(self) -> "SparkFoldsIterator":
        """Set counter to 0 and return self.

        Returns:
            Iterator for folds.

        """
        logger.debug("Creating folds iterator")

        self._curr_idx = 0

        return self

    def __next__(self) -> Tuple[SparkDataset, SparkDataset, SparkDataset]:
        """Define how to get next object.

        Returns:
            None, train dataset, validation dataset.

        """
        logger.debug(f"The next valid fold num: {self._curr_idx}")

        if self._curr_idx == self.n_folds:
            logger.debug("No more folds to continue, stopping iterations")
            raise StopIteration

        full_ds, train_part_ds, valid_part_ds = self._split_by_fold(self._curr_idx)
        self._curr_idx += 1

        return full_ds, train_part_ds, valid_part_ds

    def get_validation_data(self) -> SparkDataset:
        """Just return train dataset.

        Returns:
            Whole train dataset.

        """
        return self.train

    def convert_to_holdout_iterator(self) -> SparkHoldoutIterator:
        """Convert iterator to hold-out-iterator.

        Fold 0 is used for validation, everything else is used for training.

        Returns:
            new hold-out-iterator.

        """
        return SparkHoldoutIterator(self.train, self.input_roles)

    def combine_val_preds(self, val_preds: Sequence[SparkDataFrame], include_train: bool = False) -> SparkDataFrame:
        assert len(val_preds) > 0

        if len(val_preds) == 1:
            return val_preds[0]

        full_val_preds = functools.reduce(lambda x, y: x.unionByName(y), val_preds)

        # This transformer is necessary to prevent uneven distribution of partitions after coalescing
        # It assumes that the dataset sent in had been splitted before union corresponding to folds
        # The transformer is very specific and shouldn't be used anywhere else
        full_val_preds = BalancedUnionPartitionsCoalescerTransformer().transform(full_val_preds)

        return full_val_preds



import logging
from typing import Optional, Dict, List

import numpy as np
import pandas as pd
from pyspark.ml import Transformer
from pyspark.ml.feature import QuantileDiscretizer, Bucketizer
from pyspark.sql import functions as F
from pyspark.sql.pandas.functions import pandas_udf, PandasUDFType
from pyspark.sql.types import FloatType, IntegerType
from lightautoml.dataset.base import RolesDict

from lightautoml.dataset.roles import ColumnRole, NumericRole, CategoryRole
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.utils import SparkDataFrame
from sparklightautoml.mlwriters import CommonPickleMLReadable, CommonPickleMLWritable
from sparklightautoml.transformers.base import SparkBaseEstimator, SparkBaseTransformer, ObsoleteSparkTransformer
from lightautoml.transformers.numeric import numeric_check


logger = logging.getLogger(__name__)


class SparkNaNFlagsEstimator(SparkBaseEstimator):
    """Estimator that calculate nan rate for input columns and build :class:`~sparklightautoml.transformers.numeric.SparkNaNFlagsTransformer`."""

    _fit_checks = (numeric_check,)
    _transform_checks = ()
    # TODO: the value is copied from the corresponding LAMA transformer.
    # TODO: it is better to be taken from shared module as a string constant
    _fname_prefix = "nanflg"

    def __init__(
        self,
        input_cols: List[str],
        input_roles: Dict[str, ColumnRole],
        do_replace_columns: bool = False,
        nan_rate: float = 0.005,
    ):
        """

        Args:
            nan_rate: Nan rate cutoff.

        """
        super().__init__(
            input_cols, input_roles, do_replace_columns=do_replace_columns, output_role=NumericRole(np.float32)
        )
        self._nan_rate = nan_rate
        self._nan_cols: Optional[str] = None
        self.set(self.outputRoles, dict())
        self.set(self.outputCols, [])
        # self._features: Optional[List[str]] = None

    def _fit(self, sdf: SparkDataFrame) -> "Transformer":

        row = sdf.select([F.mean(F.isnan(c).astype(FloatType())).alias(c) for c in self.getInputCols()]).first()

        self._nan_cols = [
            f"{self._fname_prefix}__{col}"
            for col, col_nan_rate in row.asDict(True).items()
            if col_nan_rate > self._nan_rate
        ]

        self.set(self.outputCols, self._nan_cols)
        self.set(self.outputRoles, {c: self._output_role for c in self._nan_cols})

        return SparkNaNFlagsTransformer(
            input_cols=self.getInputCols(),
            input_roles=self.getInputRoles(),
            output_cols=self.getOutputCols(),
            output_roles=self.getOutputRoles(),
            do_replace_columns=self.getDoReplaceColumns(),
        )


class SparkNaNFlagsTransformer(SparkBaseTransformer, CommonPickleMLWritable, CommonPickleMLReadable):
    """Adds columns with nan flags (0 or 1) for input columns."""

    _fit_checks = (numeric_check,)
    _transform_checks = ()
    # TODO: the value is copied from the corresponding LAMA transformer.
    # TODO: it is better to be taken from shared module as a string constant
    _fname_prefix = "nanflg"

    # def __init__(self,
    #              input_cols: List[str],
    #              input_roles: RolesDict,
    #              output_cols: List[str],
    #              output_roles: RolesDict,):
    #     super().__init__(
    #         input_cols=input_cols,
    #         output_cols=output_cols,
    #         input_roles=input_roles,
    #         output_roles=output_roles,
    #         do_replace_columns=False)
    #     self._nan_cols = nan_cols

    def _transform(self, sdf: SparkDataFrame) -> SparkDataFrame:

        new_cols = [
            F.isnan(in_c).astype(FloatType()).alias(out_c)
            for in_c, out_c in zip(self.getInputCols(), self.getOutputCols())
        ]

        out_sdf = self._make_output_df(sdf, new_cols)

        return out_sdf


class SparkFillInfTransformer(SparkBaseTransformer, CommonPickleMLWritable, CommonPickleMLReadable):
    """Transformer that replace inf values to np.nan values in input columns."""

    _fit_checks = (numeric_check,)
    _transform_checks = ()
    _fname_prefix = "fillinf"

    def __init__(self, input_cols: List[str], input_roles: RolesDict, do_replace_columns=False):
        output_cols = [f"{self._fname_prefix}__{feat}" for feat in input_cols]
        super().__init__(
            input_cols=input_cols,
            output_cols=output_cols,
            input_roles=input_roles,
            output_roles={f: NumericRole(np.float32) for f in output_cols},
            do_replace_columns=do_replace_columns,
        )

    def _transform(self, df: SparkDataFrame) -> SparkDataFrame:
        def is_inf(col: str):
            return F.col(col).isin([F.lit("+Infinity").cast("double"), F.lit("-Infinity").cast("double")])

        new_cols = [
            F.when(is_inf(i), np.nan).otherwise(F.col(i)).alias(f"{self._fname_prefix}__{i}")
            for i in self.getInputCols()
        ]

        out_df = self._make_output_df(df, cols_to_add=new_cols)

        return out_df


class SparkFillnaMedianEstimator(SparkBaseEstimator):
    """Fillna with median."""

    _fit_checks = (numeric_check,)
    _transform_checks = ()
    _fname_prefix = "fillnamed"

    def __init__(
        self,
        input_cols: List[str],
        input_roles: Dict[str, ColumnRole],
        do_replace_columns: bool = False,
        subsample: int = 1_000_000,
        seed: int = 42,
    ):
        super().__init__(
            input_cols, input_roles, do_replace_columns=do_replace_columns, output_role=NumericRole(np.float32)
        )
        self._meds: Optional[Dict[str, float]] = None
        self._subsample = subsample
        self._seed = seed

    def _fit(self, sdf: SparkDataFrame) -> Transformer:
        """Approximately estimates medians.

        Args:
            dataset: SparkDataFrame with numerical features.

        Returns:
            Spark MLlib Transformer

        """

        logger.debug(f"Starting to fit estimator {self}")

        total_number = sdf.count()
        if self._subsample > total_number:
            fraction = 1.0
        else:
            fraction = self._subsample / total_number
        sdf = sdf.sample(fraction=fraction, seed=self._seed)
        logger.debug(f"Sample size: {sdf.count()}")

        row = (
            sdf.select([F.percentile_approx(c, 0.5).alias(c) for c in self.getInputCols()])
            .select([F.when(F.isnan(c), 0).otherwise(F.col(c)).alias(c) for c in self.getInputCols()])
            .first()
        )

        self._meds = row.asDict()

        logger.debug(f"Finished to fit estimator {self}")

        return SparkFillnaMedianTransformer(
            input_cols=self.getInputCols(),
            output_cols=self.getOutputCols(),
            input_roles=self.getInputRoles(),
            output_roles=self.getOutputRoles(),
            meds=self._meds,
            do_replace_columns=self.getDoReplaceColumns(),
        )


class SparkFillnaMedianTransformer(SparkBaseTransformer, CommonPickleMLWritable, CommonPickleMLReadable):
    """Fillna with median."""

    _fit_checks = (numeric_check,)
    _transform_checks = ()
    _fname_prefix = "fillnamed"

    def __init__(
        self,
        input_cols: List[str],
        output_cols: List[str],
        input_roles: RolesDict,
        output_roles: RolesDict,
        meds: Dict,
        do_replace_columns: bool = False,
    ):
        super().__init__(
            input_cols=input_cols,
            output_cols=output_cols,
            input_roles=input_roles,
            output_roles=output_roles,
            do_replace_columns=do_replace_columns,
        )
        self._meds = meds

    def _transform(self, sdf: SparkDataFrame) -> SparkDataFrame:
        """Transform - fillna with medians.

        Args:
            dataset: SparkDataFrame of numerical features

        Returns:
            SparkDataFrame with replaced NaN with medians

        """

        new_cols = [
            F.when(F.isnan(c), self._meds[c]).otherwise(F.col(c)).alias(f"{self._fname_prefix}__{c}")
            for c in self.getInputCols()
        ]

        out_sdf = self._make_output_df(sdf, new_cols)

        return out_sdf


class SparkLogOddsTransformer(SparkBaseTransformer, CommonPickleMLWritable, CommonPickleMLReadable):
    """Convert probs to logodds."""

    _fit_checks = (numeric_check,)
    _transform_checks = ()
    _fname_prefix = "logodds"

    _can_unwind_parents = False

    def __init__(self, input_cols: List[str], input_roles: RolesDict, do_replace_columns=False):
        out_cols = [f"{self._fname_prefix}__{feat}" for feat in input_cols]
        super().__init__(
            input_cols=input_cols,
            output_cols=out_cols,
            input_roles=input_roles,
            output_roles={f: NumericRole(np.float32) for f in out_cols},
            do_replace_columns=do_replace_columns,
        )

    def _transform(self, sdf: SparkDataFrame) -> SparkDataFrame:
        """Transform - convert num values to logodds.

        Args:
            dataset: SparkDataFrame dataset of categorical features.

        Returns:
            SparkDataFrame with encoded labels.

        """
        new_cols = []
        for i in self.getInputCols():
            col = F.when(F.col(i) < 1e-7, 1e-7).when(F.col(i) > 1 - 1e-7, 1 - 1e-7).otherwise(F.col(i))
            col = F.log(col / (F.lit(1) - col))
            new_cols.append(col.alias(f"{self._fname_prefix}__{i}"))

        out_sdf = self._make_output_df(sdf, new_cols)

        return out_sdf


class SparkStandardScalerEstimator(SparkBaseEstimator):
    """Classic StandardScaler."""

    _fit_checks = (numeric_check,)
    _transform_checks = ()
    _fname_prefix = "scaler"

    def __init__(self, input_cols: List[str], input_roles: Dict[str, ColumnRole], do_replace_columns: bool = False):
        super().__init__(
            input_cols, input_roles, do_replace_columns=do_replace_columns, output_role=NumericRole(np.float32)
        )
        self._means_and_stds: Optional[Dict[str, float]] = None

    def _fit(self, sdf: SparkDataFrame) -> Transformer:
        """Estimate means and stds.

        Args:
            sdf: SparkDataFrame of categorical features.

        Returns:
            StandardScalerTransformer instance

        """

        means = [F.mean(c).alias(f"mean_{c}") for c in self.getInputCols()]
        stds = [
            F.when(F.stddev(c) == 0, 1).when(F.isnan(F.stddev(c)), 1).otherwise(F.stddev(c)).alias(f"std_{c}")
            for c in self.getInputCols()
        ]

        self._means_and_stds = sdf.select(means + stds).first().asDict()

        return SparkStandardScalerTransformer(
            input_cols=self.getInputCols(),
            output_cols=self.getOutputCols(),
            input_roles=self.getInputRoles(),
            output_roles=self.getOutputRoles(),
            means_and_stds=self._means_and_stds,
            do_replace_columns=self.getDoReplaceColumns(),
        )


class SparkStandardScalerTransformer(SparkBaseTransformer, CommonPickleMLWritable, CommonPickleMLReadable):
    """Classic StandardScaler."""

    _fit_checks = (numeric_check,)
    _transform_checks = ()
    _fname_prefix = "scaler"

    def __init__(
        self,
        input_cols: List[str],
        output_cols: List[str],
        input_roles: RolesDict,
        output_roles: RolesDict,
        means_and_stds: Dict,
        do_replace_columns: bool = False,
    ):
        super().__init__(
            input_cols=input_cols,
            output_cols=output_cols,
            input_roles=input_roles,
            output_roles=output_roles,
            do_replace_columns=do_replace_columns,
        )
        self._means_and_stds = means_and_stds

    def _transform(self, sdf: SparkDataFrame) -> SparkDataFrame:
        """Scale test data.

        Args:
            sdf: SparkDataFrame of numeric features.

        Returns:
            SparkDataFrame with encoded labels.

        """

        new_cols = []
        for c in self.getInputCols():
            col = (F.col(c) - self._means_and_stds[f"mean_{c}"]) / F.lit(self._means_and_stds[f"std_{c}"])
            new_cols.append(col.alias(f"{self._fname_prefix}__{c}"))

        out_sdf = self._make_output_df(sdf, new_cols)

        return out_sdf


class SparkQuantileBinningEstimator(SparkBaseEstimator):
    """Discretization of numeric features by quantiles."""

    _fit_checks = (numeric_check,)
    _transform_checks = ()
    _fname_prefix = "qntl"

    def __init__(
        self,
        input_cols: List[str],
        input_roles: Dict[str, ColumnRole],
        do_replace_columns: bool = False,
        nbins: int = 10,
    ):
        super().__init__(
            input_cols,
            input_roles,
            do_replace_columns=do_replace_columns,
            output_role=CategoryRole(np.int32, label_encoded=True),
        )
        self._nbins = nbins
        self._bucketizer = None

    def _fit(self, sdf: SparkDataFrame) -> Transformer:
        qdisc = QuantileDiscretizer(
            numBucketsArray=[self._nbins for _ in self.getInputCols()],
            handleInvalid="keep",
            inputCols=self.getInputCols(),
            outputCols=self.getOutputCols(),
        )

        self._bucketizer = qdisc.fit(sdf)

        return SparkQuantileBinningTransformer(
            self._nbins,
            self._bucketizer,
            input_cols=self.getInputCols(),
            input_roles=self.getInputRoles(),
            output_cols=self.getOutputCols(),
            output_roles=self.getOutputRoles(),
            do_replace_columns=self.getDoReplaceColumns(),
        )


class SparkQuantileBinningTransformer(SparkBaseTransformer, CommonPickleMLWritable, CommonPickleMLReadable):
    """Adds column with quantile bin number of input columns.

    Quantile bin number of column value calculated by `QuantileDiscretizer` in SparkQuantileBinningEstimator.
    """

    _fit_checks = (numeric_check,)
    _transform_checks = ()
    _fname_prefix = "qntl"

    def __init__(
        self,
        bins,
        bucketizer,
        input_cols: List[str],
        output_cols: List[str],
        input_roles: RolesDict,
        output_roles: RolesDict,
        do_replace_columns: bool = False,
    ):
        super().__init__(
            input_cols=input_cols,
            output_cols=output_cols,
            input_roles=input_roles,
            output_roles=output_roles,
            do_replace_columns=do_replace_columns,
        )
        self._bins = bins
        self._bucketizer = bucketizer

    def _transform(self, sdf: SparkDataFrame) -> SparkDataFrame:
        new_cols = [
            F.when(F.col(c).astype(IntegerType()) == F.lit(self._bins), 0)
            .otherwise(F.col(c).astype(IntegerType()) + 1)
            .alias(c)
            for c in self._bucketizer.getOutputCols()
        ]

        rest_cols = [c for c in sdf.columns if c not in self._bucketizer.getOutputCols()]

        intermediate_sdf = self._bucketizer.transform(sdf).select(*rest_cols, *new_cols)

        output_sdf = self._make_output_df(intermediate_sdf, cols_to_add=[])

        return output_sdf



from importlib_metadata import version
from packaging.version import parse
from pyspark.sql.functions import countDistinct as count_distinct

if parse(version('pyspark')) >= parse('3.1.0'):
    from pyspark.ml.functions import array_to_vector
    from pyspark.sql.functions import percentile_approx
    from pyspark.sql.functions import aggregate
    from pyspark.sql.functions import transform
else:
    from pyspark import SparkContext
    from pyspark.sql import Column
    from pyspark.sql.column import _to_java_column, _to_seq, _create_column_from_literal
    from pyspark.sql.functions import countDistinct as count_distinct

    def _get_lambda_parameters(f):
        import inspect

        signature = inspect.signature(f)
        parameters = signature.parameters.values()

        # We should exclude functions that use
        # variable args and keyword argnames
        # as well as keyword only args
        supported_parameter_types = {
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            inspect.Parameter.POSITIONAL_ONLY,
        }

        # Validate that
        # function arity is between 1 and 3
        if not (1 <= len(parameters) <= 3):
            raise ValueError(
                "f should take between 1 and 3 arguments, but provided function takes {}".format(
                    len(parameters)
                )
            )

        # and all arguments can be used as positional
        if not all(p.kind in supported_parameter_types for p in parameters):
            raise ValueError(
                "f should use only POSITIONAL or POSITIONAL OR KEYWORD arguments"
            )

        return parameters

    def _unresolved_named_lambda_variable(*name_parts):
        """
        Create `o.a.s.sql.expressions.UnresolvedNamedLambdaVariable`,
        convert it to o.s.sql.Column and wrap in Python `Column`

        Parameters
        ----------
        name_parts : str
        """
        sc = SparkContext._active_spark_context
        name_parts_seq = _to_seq(sc, name_parts)
        expressions = sc._jvm.org.apache.spark.sql.catalyst.expressions
        return Column(
            sc._jvm.Column(
                expressions.UnresolvedNamedLambdaVariable(name_parts_seq)
            )
        )

    def _create_lambda(f):
        """
        Create `o.a.s.sql.expressions.LambdaFunction` corresponding
        to transformation described by f

        :param f: A Python of one of the following forms:
                - (Column) -> Column: ...
                - (Column, Column) -> Column: ...
                - (Column, Column, Column) -> Column: ...
        """
        parameters = _get_lambda_parameters(f)

        sc = SparkContext._active_spark_context
        expressions = sc._jvm.org.apache.spark.sql.catalyst.expressions

        argnames = ["x", "y", "z"]
        args = [
            _unresolved_named_lambda_variable(arg) for arg in argnames[: len(parameters)]
        ]

        result = f(*args)

        if not isinstance(result, Column):
            raise ValueError("f should return Column, got {}".format(type(result)))

        jexpr = result._jc.expr()
        jargs = _to_seq(sc, [arg._jc.expr() for arg in args])

        return expressions.LambdaFunction(jexpr, jargs, False)

    def _invoke_higher_order_function(name, cols, funs):
        sc = SparkContext._active_spark_context
        expressions = sc._jvm.org.apache.spark.sql.catalyst.expressions
        expr = getattr(expressions, name)

        jcols = [_to_java_column(col).expr() for col in cols]
        jfuns = [_create_lambda(f) for f in funs]

        return Column(sc._jvm.Column(expr(*jcols + jfuns)))

    def array_to_vector(col):
        sc = SparkContext._active_spark_context
        return Column(
            sc._jvm.org.apache.spark.sql.lightautoml.functions.array_to_vector(_to_java_column(col)))


    def percentile_approx(col, percentage, accuracy=10000):
        sc = SparkContext._active_spark_context

        if isinstance(percentage, (list, tuple)):
            # A local list
            percentage = sc._jvm.functions.array(_to_seq(sc, [
                _create_column_from_literal(x) for x in percentage
            ]))
        elif isinstance(percentage, Column):
            # Already a Column
            percentage = _to_java_column(percentage)
        else:
            # Probably scalar
            percentage = _create_column_from_literal(percentage)

        accuracy = (
            _to_java_column(accuracy) if isinstance(accuracy, Column)
            else _create_column_from_literal(accuracy)
        )

        return Column(sc._jvm.org.apache.spark.sql.lightautoml.functions.percentile_approx(
            _to_java_column(col), percentage, accuracy
        ))

    def aggregate(col, initialValue, merge, finish=None):
        if finish is not None:
            return _invoke_higher_order_function(
                "ArrayAggregate",
                [col, initialValue],
                [merge, finish]
            )

        else:
            return _invoke_higher_order_function(
                "ArrayAggregate",
                [col, initialValue],
                [merge]
            )

    def transform(col, f):
        return _invoke_higher_order_function("ArrayTransform", [col], [f])

array_to_vector = array_to_vector
count_distinct = count_distinct
percentile_approx = percentile_approx
aggregate = aggregate
transform = transform



from typing import Optional, Union, List

import numpy as np
import pandas as pd
from pyspark.ml import Pipeline
from pyspark.sql import functions as F
from pyspark.sql.types import IntegerType, StructField, StructType

from lightautoml.dataset.roles import CategoryRole
from lightautoml.reader.guess_roles import calc_ginis, RolesDict
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.transformers.base import SparkChangeRolesTransformer
from sparklightautoml.transformers.categorical import (
    SparkLabelEncoderEstimator,
    SparkFreqEncoderEstimator,
    SparkOrdinalEncoderEstimator,
    SparkTargetEncoderEstimator,
    SparkMulticlassTargetEncoderEstimator,
)
from sparklightautoml.transformers.numeric import SparkQuantileBinningEstimator
from lightautoml.transformers.categorical import MultiClassTargetEncoder


def get_gini_func(target_col: str):
    """Returns generator that take iterator by pandas dataframes and yield dataframes with calculated ginis.

    Args:
        target_col (str): target column to calc ginis
    """

    def gini_func(iterator):
        for pdf in iterator:
            target = pdf[target_col].to_numpy()
            data = pdf.drop(target_col, axis=1)
            cols = data.columns
            data = data.to_numpy()
            scores = calc_ginis(data, target, None)
            yield pd.DataFrame(data=[scores], columns=cols)

    return gini_func


def get_score_from_pipe(train: SparkDataset, pipe: Optional[Pipeline] = None) -> np.ndarray:
    """Get normalized gini index from pipeline.

    Args:
        train:  np.ndarray.
        target: np.ndarray.
        pipe: LAMLTransformer.
        empty_slice: np.ndarray.

    Returns:
        np.ndarray.

    """

    if pipe is not None:
        pipeline_model = pipe.fit(train.data)
        df = pipeline_model.transform(train.data)
        last_stage = pipeline_model.stages[-1]

        sdf = df.select(SparkDataset.ID_COLUMN, train.target_column, *last_stage.getOutputCols())

        train = train.empty()
        train.set_data(sdf, features=last_stage.getOutputCols(), roles=last_stage.getOutputRoles())
    else:
        sdf = train.data

    gini_func = get_gini_func(train.target_column)

    # schema without target column
    output_schema = (train.data.select(*train.features)).schema
    # need to set True
    output_schema = StructType([StructField(f.name, f.dataType, True) for f in output_schema.fields])

    mean_scores = (
        (
            train.data.select(*train.features, train.target_column)
            .mapInPandas(gini_func, output_schema)
            .select([F.mean(c).alias(c) for c in train.features])
        )
        .toPandas()
        .values.flatten()
    )

    return mean_scores


def get_numeric_roles_stat(
    train: SparkDataset,
    subsample: Optional[Union[float, int]] = 100000,
    random_state: int = 42,
    manual_roles: Optional[RolesDict] = None,
) -> pd.DataFrame:
    """Calculate statistics about different encodings performances.

    We need it to calculate rules about advanced roles guessing.
    Only for numeric data.

    Args:
        train: Dataset.
        subsample: size of subsample.
        random_state: int.
        manual_roles: Dict.
        n_jobs: int.

    Returns:
        DataFrame.

    """
    if manual_roles is None:
        manual_roles = {}

    roles_to_identify = []
    roles = []
    flg_manual_set = []
    # check for train dtypes
    for f in train.features:
        role = train.roles[f]
        if role.name == "Numeric":
            roles_to_identify.append(f)
            roles.append(role)
            flg_manual_set.append(f in manual_roles)

    res = pd.DataFrame(
        columns=[
            "flg_manual",
            "unique",
            "unique_rate",
            "top_freq_values",
            "raw_scores",
            "binned_scores",
            "encoded_scores",
            "freq_scores",
            "nan_rate",
        ],
        index=roles_to_identify,
    )
    res["flg_manual"] = flg_manual_set

    if len(roles_to_identify) == 0:
        return res

    sdf = train.data.select(SparkDataset.ID_COLUMN, train.folds_column, train.target_column, *roles_to_identify)

    total_number = sdf.count()
    if subsample is not None:
        if subsample > total_number:
            fraction = 1.0
        else:
            fraction = subsample / total_number
            total_number = subsample
        sdf = sdf.sample(fraction=fraction, seed=random_state)

    train = train.empty()
    train.set_data(sdf, roles_to_identify, roles)

    assert train.folds is not None

    # if train.folds is None:
    #     train.folds = set_sklearn_folds(train.task, train.target, cv=5, random_state=42, group=train.group)

    data, target = train.data, train.target

    # check task specific
    if train.task.name == "multiclass":
        encoder = SparkMulticlassTargetEncoderEstimator
    else:
        encoder = SparkTargetEncoderEstimator

    # s3d = data.shape + (-1,)
    # empty_slice = np.isnan(data)

    # check scores as is
    res["raw_scores"] = get_score_from_pipe(train)

    # # check unique values
    # sub_select_columns = []
    # top_select_columns = []
    # for f in train.features:
    #     sub_select_columns.append(F.count(F.when(~F.isnan(F.col(f)), F.col(f))).over(Window.partitionBy(F.col(f))).alias(f'{f}_count_values'))
    #     top_select_columns.append(F.max(F.col(f'{f}_count_values')).alias(f'{f}_max_count_values'))
    #     top_select_columns.append(F.count_distinct(F.when(~F.isnan(F.col(f)), F.col(f))).alias(f'{f}_count_distinct'))
    # df = train.data.select(*train.features, *sub_select_columns)
    # unique_values_stat: Dict = df.select(*top_select_columns).first().asDict()
    #
    # # max of frequency of unique values in every column
    # res["top_freq_values"] = np.array([unique_values_stat[f'{f}_max_count_values'] for f in train.features])
    # # how many unique values in every column
    # res["unique"] = np.array([unique_values_stat[f'{f}_count_distinct'] for f in train.features])
    # res["unique_rate"] = res["unique"] / total_number

    # check binned categorical score
    quantile_binning = SparkQuantileBinningEstimator(input_cols=train.features, input_roles=train.roles)
    target_encoder = encoder(
        input_cols=quantile_binning.getOutputCols(),
        input_roles=quantile_binning.getOutputRoles(),
        task_name=train.task.name,
        folds_column=train.folds_column,
        target_column=train.target_column,
        do_replace_columns=True,
    )
    trf = Pipeline(stages=[quantile_binning, target_encoder])
    res["binned_scores"] = get_score_from_pipe(train, pipe=trf)

    # check label encoded scores
    change_roles = SparkChangeRolesTransformer(
        input_cols=train.features, input_roles=train.roles, role=CategoryRole(np.float32)
    )
    label_encoder = SparkLabelEncoderEstimator(
        input_cols=change_roles.getOutputCols(), input_roles=change_roles.getOutputRoles(), random_state=random_state
    )
    target_encoder = encoder(
        input_cols=label_encoder.getOutputCols(),
        input_roles=label_encoder.getOutputRoles(),
        task_name=train.task.name,
        folds_column=train.folds_column,
        target_column=train.target_column,
        do_replace_columns=True,
    )
    trf = Pipeline(stages=[change_roles, label_encoder, target_encoder])
    res["encoded_scores"] = get_score_from_pipe(train, pipe=trf)

    # check frequency encoding
    change_roles = SparkChangeRolesTransformer(
        input_cols=train.features, input_roles=train.roles, role=CategoryRole(np.float32)
    )
    freq_encoder = SparkFreqEncoderEstimator(
        input_cols=change_roles.getOutputCols(), input_roles=change_roles.getOutputRoles()
    )
    trf = Pipeline(stages=[change_roles, freq_encoder])
    res["freq_scores"] = get_score_from_pipe(train, pipe=trf)

    nan_rate_cols = [F.mean(F.isnan(F.col(feat)).astype(IntegerType())).alias(feat) for feat in train.features]
    res["nan_rate"] = train.data.select(nan_rate_cols).toPandas().values.flatten()

    return res


def get_category_roles_stat(
    train: SparkDataset, subsample: Optional[Union[float, int]] = 100000, random_state: int = 42
):
    """Search for optimal processing of categorical values.

    Categorical means defined by user or object types.

    Args:
        train: Dataset.
        subsample: size of subsample.
        random_state: seed of random numbers generator.
        n_jobs: number of jobs.

    Returns:
        result.

    """

    roles_to_identify = []

    dtypes = []

    # check for train dtypes
    roles = []
    for f in train.features:
        role = train.roles[f]
        if role.name == "Category" and role.encoding_type == "auto":
            roles_to_identify.append(f)
            roles.append(role)
            dtypes.append(role.dtype)

    res = pd.DataFrame(
        columns=[
            "unique",
            "top_freq_values",
            "dtype",
            "encoded_scores",
            "freq_scores",
            "ord_scores",
        ],
        index=roles_to_identify,
    )

    res["dtype"] = dtypes

    if len(roles_to_identify) == 0:
        return res

    sdf = train.data.select(SparkDataset.ID_COLUMN, train.folds_column, train.target_column, *roles_to_identify)

    if subsample is not None:
        total_number = sdf.count()
        if subsample > total_number:
            fraction = 1.0
        else:
            fraction = subsample / total_number
        sdf = sdf.sample(fraction=fraction, seed=random_state)

    train = train.empty()
    train.set_data(sdf, roles_to_identify, roles)

    assert train.folds is not None

    # check task specific
    if train.task.name == "multiclass":
        encoder = MultiClassTargetEncoder
    else:
        encoder = SparkTargetEncoderEstimator

    # check label encoded scores
    label_encoder = SparkLabelEncoderEstimator(
        input_cols=train.features, input_roles=train.roles, random_state=random_state
    )
    target_encoder = encoder(
        input_cols=label_encoder.getOutputCols(),
        input_roles=label_encoder.getOutputRoles(),
        task_name=train.task.name,
        folds_column=train.folds_column,
        target_column=train.target_column,
        do_replace_columns=True,
    )
    trf = Pipeline(stages=[label_encoder, target_encoder])
    res["encoded_scores"] = get_score_from_pipe(train, pipe=trf)

    # check frequency encoding
    trf = SparkFreqEncoderEstimator(input_cols=train.features, input_roles=train.roles, do_replace_columns=True)
    trf = Pipeline(stages=[trf])
    res["freq_scores"] = get_score_from_pipe(train, pipe=trf)

    # check ordinal encoding
    trf = SparkOrdinalEncoderEstimator(input_cols=train.features, input_roles=train.roles, random_state=random_state)
    trf = Pipeline(stages=[trf])
    res["ord_scores"] = get_score_from_pipe(train, pipe=trf)

    return res


def get_null_scores(
    train: SparkDataset,
    feats: Optional[List[str]] = None,
    subsample: Optional[Union[float, int]] = 100000,
    random_state: int = 42,
) -> pd.Series:
    """Get null scores.

    Args:
        train: Dataset
        feats: list of features.
        subsample: size of subsample.
        random_state: seed of random numbers generator.

    Returns:
        Series.

    """
    roles = train.roles
    sdf = train.data.select(SparkDataset.ID_COLUMN, train.folds_column, train.target_column, *feats)

    if subsample is not None:
        total_number = sdf.count()
        if subsample > total_number:
            fraction = 1.0
        else:
            fraction = subsample / total_number
        sdf = sdf.sample(fraction=fraction, seed=random_state)

    train = train.empty()
    train.set_data(sdf, feats, [roles[f] for f in feats])

    train.data.cache()
    size = train.data.count()
    notnan = (
        train.data.select([F.sum(F.isnull(feat).astype(IntegerType())).alias(feat) for feat in train.features])
        .first()
        .asDict()
    )

    notnan_cols = [feat for feat, cnt in notnan.items() if cnt != size and cnt != 0]

    if notnan_cols:
        empty_slice_cols = [F.when(F.isnull(F.col(feat)), 1.0).otherwise(0.0).alias(feat) for feat in notnan_cols]

        gini_func = get_gini_func(train.target_column)
        sdf = train.data.select(SparkDataset.ID_COLUMN, train.target_column, *empty_slice_cols)

        output_schema = sdf.select(SparkDataset.ID_COLUMN, *notnan_cols).schema

        mean_scores = (
            (sdf.mapInPandas(gini_func, output_schema).select([F.mean(c).alias(c) for c in notnan_cols]))
            .first()
            .asDict()
        )
    else:
        mean_scores = {}

    scores = [mean_scores[feat] if feat in mean_scores else 0.0 for feat in train.features]

    train.data.unpersist()

    res = pd.Series(scores, index=train.features, name="max_score")

    return res



from typing import Union, Tuple, Optional, Sequence

from lightautoml.dataset.base import RolesDict
from lightautoml.ml_algo.tuning.base import ParamsTuner, DefaultTuner
from lightautoml.pipelines.ml.nested_ml_pipe import (
    NestedTabularMLPipeline as LAMANestedTabularMLPipeline,
    NestedTabularMLAlgo,
)
from lightautoml.pipelines.selection.base import SelectionPipeline
from sparklightautoml.ml_algo.base import SparkTabularMLAlgo
from sparklightautoml.pipelines.features.base import FeaturesPipeline, SparkFeaturesPipeline
from sparklightautoml.pipelines.ml.base import SparkMLPipeline


class SparkNestedTabularMLPipeline(SparkMLPipeline, LAMANestedTabularMLPipeline):
    """
    Same as NestedTabularMLPipeline of LAMA, but redefines a couple of methods via SparkMLPipelineMixin
    """

    def __init__(
        self,
        cacher_key: str,
        ml_algos: Sequence[Union[SparkTabularMLAlgo, Tuple[SparkTabularMLAlgo, ParamsTuner]]],
        force_calc: Union[bool, Sequence[bool]] = True,
        pre_selection: Optional[SelectionPipeline] = None,
        features_pipeline: Optional[SparkFeaturesPipeline] = None,
        post_selection: Optional[SelectionPipeline] = None,
        cv: int = 1,
        n_folds: Optional[int] = None,
        inner_tune: bool = False,
        refit_tuner: bool = False,
    ):
        if cv > 1:
            new_ml_algos = []

            for n, mt_pair in enumerate(ml_algos):

                try:
                    mod, tuner = mt_pair
                except (TypeError, ValueError):
                    mod, tuner = mt_pair, DefaultTuner()

                if inner_tune:
                    new_ml_algos.append(NestedTabularMLAlgo(mod, tuner, refit_tuner, cv, n_folds))
                else:
                    new_ml_algos.append((NestedTabularMLAlgo(mod, None, True, cv, n_folds), tuner))

            ml_algos = new_ml_algos
        super().__init__(cacher_key, ml_algos, force_calc, pre_selection, features_pipeline, post_selection)



from copy import copy, deepcopy
from typing import List, Optional, Dict, Tuple, Any, Union

import gensim
import numpy as np
import pandas as pd
from pyspark.ml import Pipeline, PipelineModel
from pyspark.ml.feature import RegexTokenizer as PysparkRegexTokenizer
from pyspark.ml.feature import Tokenizer as PysparkTokenizer, CountVectorizer, IDF, Normalizer
from pyspark.sql.functions import concat_ws
from pyspark.sql.pandas.functions import pandas_udf, PandasUDFType
from sklearn.base import TransformerMixin

from lightautoml.dataset.roles import TextRole
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.dataset.roles import NumericVectorOrArrayRole
from sparklightautoml.transformers.base import ObsoleteSparkTransformer
from lightautoml.text.dl_transformers import DLTransformer
from lightautoml.text.tokenizer import BaseTokenizer
from lightautoml.text.tokenizer import SimpleEnTokenizer
from lightautoml.transformers.text import TunableTransformer, model_by_name, logger
from lightautoml.transformers.text import text_check


class TfidfTextTransformer(ObsoleteSparkTransformer, TunableTransformer):
    """Simple Tfidf vectorizer."""

    _fit_checks = (text_check,)
    _transform_checks = ()
    _fname_prefix = "tfidf"
    _default_params = {"min_df": 5, "max_df": 1.0, "max_features": 30_000, "dtype": np.float32, "normalization": 2.0}

    # These properties are not supported
    # cause there is no analogues in Spark ML
    # "ngram_range": (1, 1),
    # "analyzer": "word",

    @property
    def features(self) -> List[str]:
        """Features list."""

        return self._features

    def __init__(
        self,
        default_params: Optional[dict] = None,
        freeze_defaults: bool = True,
        subs: Optional[float] = None,
        random_state: int = 42,
    ):
        """

        Args:
            default_params: algo hyperparams.
            freeze_defaults: Flag.
            subs: Subsample to calculate freqs. If ``None`` - full data.
            random_state: Random state to take subsample.

        Note:
            The behaviour of `freeze_defaults`:

            - ``True`` :  params may be rewritten depending on dataset.
            - ``False``:  params may be changed only
              manually or with tuning.

        """
        super().__init__(default_params, freeze_defaults)
        self.subs = subs
        self.random_state = random_state
        self.idf_columns_pipelines: Optional[Dict[str, Tuple[PipelineModel, str, int]]] = None
        self.vocab_size: Optional[int] = None

    def init_params_on_input(self, dataset: SparkDataset) -> dict:
        """Get transformer parameters depending on dataset parameters.

        Args:
            dataset: Dataset used for model parmaeters initialization.

        Returns:
            Parameters of model.

        """

        # TODO: use features_num
        suggested_params = copy(self.default_params)
        if self.freeze_defaults:
            # if user change defaults manually - keep it
            return suggested_params

        # TODO: decide later what to do with this part
        # rows_num = len(dataset.data)
        # if rows_num > 50_000:
        #     suggested_params["min_df"] = 25

        return suggested_params

    def _fit(self, dataset: SparkDataset):
        """Fit tfidf vectorizer.

        Args:
            dataset: Pandas or Numpy dataset of text features.

        Returns:
            self.

        """

        # convert to accepted dtype and get attributes
        if self._params is None:
            self.params = self.init_params_on_input(dataset)

        sdf = dataset.data
        if self.subs:
            sdf = sdf.sample(self.subs, seed=self.random_state)
        sdf = sdf.fillna("")

        self.idf_columns_pipelines = dict()
        feats = []
        for c in dataset.features:
            # TODO: set params here from self.params
            tokenizer = PysparkTokenizer(inputCol=c, outputCol=f"{c}_words")
            count_tf = CountVectorizer(
                minDF=self.params["min_df"],
                maxDF=self.params["max_df"],
                vocabSize=self.params["max_features"],
                inputCol=tokenizer.getOutputCol(),
                outputCol=f"{c}_word_features",
            )
            out_col = f"{self._fname_prefix}__{c}"
            idf = IDF(inputCol=count_tf.getOutputCol(), outputCol=f"{c}_idf_features")

            stages = [tokenizer, count_tf, idf]
            if self.params["normalization"] and self.params["normalization"] > 0:
                norm = Normalizer(inputCol=idf.getOutputCol(), outputCol=out_col, p=self.params["normalization"])
                stages.append(norm)

            pipeline = Pipeline(stages=stages)
            tfidf_pipeline_model = pipeline.fit(sdf)

            # features = list(
            #     np.char.array([self._fname_prefix + "_"])
            #     + np.arange(count_tf.getVocabSize()).astype(str)
            #     + np.char.array(["__" + c])
            # )
            # feats.extend(features)

            self.idf_columns_pipelines[c] = (tfidf_pipeline_model, out_col, count_tf.getVocabSize())

        self._features = feats

        return self

    def _transform(self, dataset: SparkDataset) -> SparkDataset:
        """Transform text dataset to sparse tfidf representation.

        Args:
            dataset: Pandas or Numpy dataset of text features.

        Returns:
            Sparse dataset with encoded text.

        """

        sdf = dataset.data
        sdf = sdf.fillna("")

        # transform

        curr_sdf = sdf
        all_idf_features = []
        all_idf_roles = []
        for c in dataset.features:
            tfidf_model, idf_col, vocab_size = self.idf_columns_pipelines[c]

            curr_sdf = tfidf_model.transform(curr_sdf)

            role = NumericVectorOrArrayRole(
                size=vocab_size, element_col_name_template=f"{self._fname_prefix}_{{}}__{idf_col}"
            )

            all_idf_features.append(idf_col)
            all_idf_roles.append(role)

        # all_idf_features = [
        #     vector_to_array(F.col(idf_col))[i].alias(feat)
        #     for idf_col, features in idf2features.items()
        #     for i,feat in enumerate(features)
        # ]

        new_sdf = curr_sdf.select(*dataset.service_columns, *all_idf_features)

        output = dataset.empty()
        output.set_data(new_sdf, self._features, all_idf_roles)

        return output


class AutoNLPWrap(ObsoleteSparkTransformer):
    """Calculate text embeddings."""

    _fit_checks = (text_check,)
    _transform_checks = ()
    _fname_prefix = "emb"
    fasttext_params = {"vector_size": 64, "window": 3, "min_count": 1}
    _names = {"random_lstm", "random_lstm_bert", "pooled_bert", "wat", "borep"}
    _trainable = {"wat", "borep", "random_lstm"}

    @property
    def features(self) -> List[str]:
        """Features list."""
        return self._features

    def __init__(
        self,
        model_name: str,
        embedding_model: Optional[str] = None,
        cache_dir: str = "./cache_NLP",
        bert_model: Optional[str] = None,
        transformer_params: Optional[Dict] = None,
        subs: Optional[int] = None,
        multigpu: bool = False,
        random_state: int = 42,
        train_fasttext: bool = False,
        fasttext_params: Optional[Dict] = None,
        fasttext_epochs: int = 2,
        sent_scaler: Optional[str] = None,
        verbose: bool = False,
        device: Any = "0",
        **kwargs: Any,
    ):
        """

        Args:
            model_name: Method for aggregating word embeddings
              into sentence embedding.
            transformer_params: Aggregating model parameters.
            embedding_model: Word level embedding model with dict
              interface or path to gensim fasttext model.
            cache_dir: If ``None`` - do not cache transformed datasets.
            bert_model: Name of HuggingFace transformer model.
            subs: Subsample to calculate freqs. If None - full data.
            multigpu: Use Data Parallel.
            random_state: Random state to take subsample.
            train_fasttext: Train fasttext.
            fasttext_params: Fasttext init params.
            fasttext_epochs: Number of epochs to train.
            verbose: Verbosity.
            device: Torch device or str.
            **kwargs: Unused params.

        """
        if train_fasttext:
            assert model_name in self._trainable, f"If train fasstext then model must be in {self._trainable}"

        assert model_name in self._names, f"Model name must be one of {self._names}"
        self.device = device
        self.multigpu = multigpu
        self.cache_dir = cache_dir
        self.random_state = random_state
        self.subs = subs
        self.train_fasttext = train_fasttext
        self.fasttext_epochs = fasttext_epochs
        if fasttext_params is not None:
            self.fasttext_params.update(fasttext_params)
        self.sent_scaler = sent_scaler
        self.verbose = verbose
        self.model_name = model_name
        self.transformer_params = model_by_name[self.model_name]
        if transformer_params is not None:
            self.transformer_params.update(transformer_params)

        self._update_bert_model(bert_model)
        if embedding_model is not None:
            if isinstance(embedding_model, str):
                try:
                    embedding_model = gensim.models.FastText.load(embedding_model)
                except:
                    try:
                        embedding_model = gensim.models.FastText.load_fasttext_format(embedding_model)
                    except:
                        embedding_model = gensim.models.KeyedVectors.load(embedding_model)

            self.transformer_params = self._update_transformers_emb_model(self.transformer_params, embedding_model)

        else:

            self.train_fasttext = self.model_name in self._trainable

        if self.model_name == "wat":
            raise NotImplementedError("WeightedAverageTransformer is not yet supported, but it is planned")
            # self.transformer = WeightedAverageTransformer
        else:
            self.transformer = DLTransformer

        self.column_transformers: Optional[Dict[str, Tuple[TransformerMixin, NumericVectorOrArrayRole, str]]] = None

    def _update_bert_model(self, bert_model: str):
        if bert_model is not None:
            if "dataset_params" in self.transformer_params:
                self.transformer_params["dataset_params"]["model_name"] = bert_model
            if "embedding_model_params" in self.transformer_params:
                self.transformer_params["embedding_model_params"]["model_name"] = bert_model
            if "model_params" in self.transformer_params:
                self.transformer_params["model_params"]["model_name"] = bert_model
        return self

    def _update_transformers_emb_model(
        self, params: Dict, model: Any, emb_size: Optional[int] = None
    ) -> Dict[str, Any]:
        if emb_size is None:
            try:
                # Gensim checker [1]
                emb_size = model.vector_size
            except:
                try:
                    # Gensim checker[2]
                    emb_size = model.vw.vector_size
                except:
                    try:
                        # Natasha checker
                        emb_size = model[model.vocab.words[0]].shape[0]
                    except:
                        try:
                            # Dict of embeddings checker
                            emb_size = next(iter(model.values())).shape[0]
                        except:
                            raise ValueError("Unrecognized embedding dimention, please specify it in model_params")
        try:
            model = model.wv
        except:
            pass

        if self.model_name == "wat":
            params["embed_size"] = emb_size
            params["embedding_model"] = model
        elif self.model_name in {"random_lstm", "borep"}:
            params["dataset_params"]["embedding_model"] = model
            params["dataset_params"]["embed_size"] = emb_size
            params["model_params"]["embed_size"] = emb_size

        return params

    def _fit(self, dataset: SparkDataset):
        """Fit chosen transformer and create feature names.

        Args:
            dataset: Pandas or Numpy dataset of text features.

        """

        # TODO: cache should be reimplemented after the discussion
        # if self.cache_dir is not None:
        #     if not os.path.exists(self.cache_dir):
        #         os.makedirs(self.cache_dir)
        # set transformer features

        # convert to accepted dtype and get attributes
        # dataset = dataset.to_pandas()
        # df = dataset.data

        sdf = dataset.data

        # fit ...
        # TODO: replace subs with implementation in the base class
        # if self.subs is not None and df.shape[0] >= self.subs:
        #     subs = df.sample(n=self.subs, random_state=self.random_state)
        # else:
        #     subs = df

        names = []
        self.column_transformers = dict()
        for c in sdf.columns:
            transformer_params = deepcopy(self.transformer_params)

            # TODO: cannot implement embedding training using FastText this way
            # TODO: need to have a pytorch based version for training
            # if self.train_fasttext:
            #     embedding_model = gensim.models.FastText(**self.fasttext_params)
            #     common_texts = [i.split(" ") for i in subs[i].values]
            #     embedding_model.build_vocab(corpus_iterable=common_texts)
            #     embedding_model.train(
            #         corpus_iterable=common_texts,
            #         total_examples=len(common_texts),
            #         epochs=self.fasttext_epochs,
            #     )
            #     transformer_params = self._update_transformers_emb_model(transformer_params, embedding_model)

            transformer = self.transformer(
                verbose=self.verbose,
                device=self.device,
                multigpu=self.multigpu,
                **transformer_params,
            )
            emb_name = transformer.get_name()
            emb_size = transformer.get_out_shape()

            role = NumericVectorOrArrayRole(
                size=emb_size,
                element_col_name_template=f"{self._fname_prefix}_{emb_name}_{{}}__{c}",
                dtype=np.float32,
                is_vector=False,
            )

            feat_name = f"{self._fname_prefix}_{emb_name}__{c}"

            # TODO: the column transformer has to support working with dataframes,
            # TODO: e.g. there should be special Spark adapted version of this column transformer
            fitted_col_transformer = transformer.fit(sdf.select(c))

            self.column_transformers[c] = (fitted_col_transformer, role, feat_name)

            logger.info3(f"Feature {c} fitted")

            names.append(feat_name)

            # TODO: probably we should do that
            # del transformer
            # gc.collect()
            # torch.cuda.empty_cache()

        self._features = names
        return self

    # TODO: this method may be unified with what we have in AutoCVWrap and ImageFeaturesTransformer
    def _transform(self, dataset: SparkDataset) -> SparkDataset:
        """Transform tokenized dataset to text embeddings.

        Args:
            dataset: Pandas or Numpy dataset of text features.

        Returns:
            Numpy dataset with text embeddings.

        """

        sdf = dataset.data

        # transform
        new_cols = []
        roles = []
        for c in sdf.columns:
            # TODO: should be replaced with new implementation after a discussion
            # if self.cache_dir is not None:
            #     full_hash = get_textarr_hash(df[i]) + get_textarr_hash(self.dicts[i]["feats"])
            #     fname = os.path.join(self.cache_dir, full_hash + ".pkl")
            #     if os.path.exists(fname):
            #         logger.info3(f"Load saved dataset for {i}")
            #         with open(fname, "rb") as f:
            #             new_arr = pickle.load(f)
            #     else:
            #         new_arr = self.dicts[i]["transformer"].transform(df[i])
            #         with open(fname, "wb") as f:
            #             pickle.dump(new_arr, f)
            # else:
            trans, role, out_col_name = self.column_transformers[c]
            transformer_bcast = dataset.spark_session.sparkContext.broadcast(trans)

            @pandas_udf("array<float>", PandasUDFType.SCALAR)
            def calculate_embeddings(data: pd.Series) -> pd.Series:
                transformer = transformer_bcast.value
                embeds = pd.Series(list(transformer.transform(data)))
                return embeds

            new_cols.append(calculate_embeddings(c).alias(out_col_name))
            roles.append(role)

            logger.info3(f"Feature {c} transformed")

        new_sdf = sdf.select(new_cols)

        output = dataset.empty()
        output.set_data(new_sdf, self.features, roles)

        return output

    @staticmethod
    def _sentence_norm(x: np.ndarray, mode: Optional[str] = None) -> Union[np.ndarray, float]:
        """Get sentence embedding norm."""
        if mode == "l2":
            return ((x ** 2).sum(axis=1, keepdims=True)) ** 0.5
        elif mode == "l1":
            return np.abs(x).sum(axis=1, keepdims=True)
        if mode is not None:
            logger.info2("Unknown sentence scaler mode: sent_scaler={}, " "no normalization will be used".format(mode))
        return 1


class Tokenizer(ObsoleteSparkTransformer):
    _fit_checks = (text_check,)
    _transform_checks = ()
    _fname_prefix = "tokenized"

    _can_unwind_parents = False

    def __init__(self, tokenizer: BaseTokenizer = SimpleEnTokenizer()):
        """
        Args:
            tokenizer: text tokenizer.
        """
        self.tokenizer = tokenizer

    def _transform(self, dataset: SparkDataset) -> SparkDataset:

        spark_data_frame = dataset.data

        for i, column in enumerate(dataset.features):
            # PysparkTokenizer transforms strings to lowercase, do not use it
            # tokenizer = PysparkTokenizer(inputCol=column, outputCol=self._fname_prefix + "__" + column)

            tokenizer = PysparkRegexTokenizer(
                inputCol=column, outputCol=self._fname_prefix + "__" + column, toLowercase=False
            )
            tokenized = tokenizer.transform(spark_data_frame)
            spark_data_frame = tokenized
            spark_data_frame = spark_data_frame.drop(column)

        output = dataset.empty()
        output.set_data(spark_data_frame, self.features, TextRole(np.str))

        return output


class ConcatTextTransformer(ObsoleteSparkTransformer):
    _fit_checks = (text_check,)
    _transform_checks = ()
    _fname_prefix = "concated"

    _can_unwind_parents = False

    def __init__(self, special_token: str = " [SEP] "):
        """

        Args:
            special_token: Add special token between columns.

        """
        self.special_token = special_token

    def _transform(self, dataset: SparkDataset) -> SparkDataset:
        spark_data_frame = dataset.data
        spark_column_names = dataset.features

        colum_name = self._fname_prefix + "__" + "__".join(spark_column_names)
        concatExpr = concat_ws(self.special_token, *spark_column_names).alias(colum_name)
        concated = spark_data_frame.select(*dataset.service_columns, concatExpr)

        output = dataset.empty()
        output.set_data(concated, self.features, TextRole(np.str))

        return output



"""Basic classes for features generation."""
import itertools
import logging
from copy import copy
from dataclasses import dataclass
from typing import Any, Callable, cast, Set, Union, Dict
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
import toposort
from pandas import DataFrame
from pandas import Series
from pyspark.ml import Transformer, Estimator, Pipeline, PipelineModel
from pyspark.ml.param import Param, Params
from pyspark.sql import functions as F

from lightautoml.dataset.base import RolesDict, LAMLDataset
from lightautoml.dataset.roles import ColumnRole, NumericRole
from lightautoml.pipelines.features.base import FeaturesPipeline
from lightautoml.pipelines.utils import get_columns_by_role
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.pipelines.base import InputFeaturesAndRoles, OutputFeaturesAndRoles
from sparklightautoml.transformers.base import (
    SparkChangeRolesTransformer,
    ColumnsSelectorTransformer,
    DropColumnsTransformer,
)
from sparklightautoml.transformers.base import (
    SparkBaseEstimator,
    SparkBaseTransformer,
    SparkUnionTransformer,
    SparkSequentialTransformer,
    SparkEstOrTrans,
    SparkColumnsAndRoles,
)
from sparklightautoml.transformers.categorical import (
    SparkCatIntersectionsEstimator,
    SparkFreqEncoderEstimator,
    SparkLabelEncoderEstimator,
    SparkOrdinalEncoderEstimator,
    SparkMulticlassTargetEncoderEstimator,
)
from sparklightautoml.transformers.categorical import SparkTargetEncoderEstimator
from sparklightautoml.transformers.datetime import SparkBaseDiffTransformer, SparkDateSeasonsTransformer
from sparklightautoml.transformers.numeric import SparkQuantileBinningEstimator
from sparklightautoml.utils import NoOpTransformer, Cacher, EmptyCacher, warn_if_not_cached, SparkDataFrame

logger = logging.getLogger(__name__)


def build_graph(begin: SparkEstOrTrans):
    """Fill dict that represents graph of estimators and transformers

    Args:
        begin (SparkEstOrTrans): pipeline to extract graph of estimators and transformers
    """
    graph = dict()

    def find_start_end(tr: SparkEstOrTrans) -> Tuple[List[SparkEstOrTrans], List[SparkEstOrTrans]]:
        if isinstance(tr, SparkSequentialTransformer):
            se = [st_or_end for el in tr.transformers for st_or_end in find_start_end(el)]

            starts = se[0]
            ends = se[-1]
            middle = se[1:-1]

            i = 0
            while i < len(middle):
                for new_st, new_end in itertools.product(middle[i], middle[i + 1]):
                    if new_end not in graph:
                        graph[new_end] = set()
                    graph[new_end].add(new_st)
                i += 2

            return starts, ends

        elif isinstance(tr, SparkUnionTransformer):
            se = [find_start_end(el) for el in tr.transformers]
            starts = [s_el for s, _ in se for s_el in s]
            ends = [e_el for _, e in se for e_el in e]
            return starts, ends
        else:
            return [tr], [tr]

    init_starts, final_ends = find_start_end(begin)

    for st in init_starts:
        if st not in graph:
            graph[st] = set()

    return graph


@dataclass
class FittedPipe:
    sdf: SparkDataFrame
    transformer: Transformer
    roles: RolesDict


class SelectTransformer(Transformer):
    """
    Transformer that returns ``pyspark.sql.DataFrame`` with selected columns.
    """

    colsToSelect = Param(Params._dummy(), "colsToSelect", "columns to select from the dataframe")

    def __init__(self, cols_to_select: List[str]):
        """
        Args:
            cols_to_select (List[str]): List of columns to select from input dataframe
        """
        super().__init__()
        self.set(self.colsToSelect, cols_to_select)

    def getColsToSelect(self) -> List[str]:
        return self.getOrDefault(self.colsToSelect)

    def _transform(self, dataset):
        return dataset.select(self.getColsToSelect())


class SparkFeaturesPipeline(InputFeaturesAndRoles, OutputFeaturesAndRoles, FeaturesPipeline):
    """Abstract class.

    Analyze train dataset and create composite transformer
    based on subset of features.
    Instance can be interpreted like Transformer
    (look for :class:`~lightautoml.transformers.base.LAMLTransformer`)
    with delayed initialization (based on dataset metadata)
    Main method, user should define in custom pipeline is ``.create_pipeline``.
    For example, look at
    :class:`~lightautoml.pipelines.features.lgb_pipeline.LGBSimpleFeatures`.
    After FeaturePipeline instance is created, it is used like transformer
    with ``.fit_transform`` and ``.transform`` method.

    """

    def __init__(self, cacher_key: str = "default_cacher", **kwargs):
        super().__init__(**kwargs)
        self._cacher_key = cacher_key
        self.pipes: List[Callable[[SparkDataset], SparkEstOrTrans]] = [self.create_pipeline]
        self._transformer: Optional[Transformer] = None

    @property
    def transformer(self) -> Optional[Transformer]:
        return self._transformer

    def create_pipeline(self, train: SparkDataset) -> SparkEstOrTrans:
        """Analyse dataset and create composite transformer.

        Args:
            train: Dataset with train data.

        Returns:
            Composite transformer (pipeline).

        """
        raise NotImplementedError

    def fit_transform(self, train: SparkDataset) -> SparkDataset:
        """Create pipeline and then fit on train data and then transform.

        Args:
            train: Dataset with train data.n

        Returns:
            Dataset with new features.

        """
        logger.info("SparkFeaturePipeline is started")

        assert self.input_features is not None, "Input features should be provided before the fit_transform"
        assert self.input_roles is not None, "Input roles should be provided before the fit_transform"

        fitted_pipe = self._merge_pipes(train)
        self._transformer = fitted_pipe.transformer
        self._output_roles = fitted_pipe.roles

        features = train.features + self.output_features
        roles = copy(train.roles)
        roles.update(self._output_roles)
        transformed_ds = train.empty()
        transformed_ds.set_data(fitted_pipe.sdf, features, roles)

        logger.info("SparkFeaturePipeline is finished")

        return transformed_ds

    def transform(self, test: LAMLDataset) -> LAMLDataset:
        sdf = self._transformer.transform(test.data)

        roles = copy(test.roles)
        roles.update(self.output_roles)

        transformed_ds = test.empty()
        transformed_ds.set_data(sdf, self.output_features, roles)

        return transformed_ds

    def append(self, pipeline):
        if isinstance(pipeline, SparkFeaturesPipeline):
            pipeline = [pipeline]

        for _pipeline in pipeline:
            self.pipes.extend(_pipeline.pipes)

        return self

    def prepend(self, pipeline):
        if isinstance(pipeline, SparkFeaturesPipeline):
            pipeline = [pipeline]

        for _pipeline in reversed(pipeline):
            self.pipes = _pipeline.pipes + self.pipes

        return self

    def pop(self, i: int = -1) -> Optional[Callable[[SparkDataset], Estimator]]:
        if len(self.pipes) > 1:
            return self.pipes.pop(i)

    def _merge_pipes(self, data: SparkDataset) -> FittedPipe:
        fitted_pipes = []
        current_sdf = data.data
        for pipe in self.pipes:
            fp = self._optimize_and_fit(current_sdf, pipe(data))
            current_sdf = fp.sdf
            fitted_pipes.append(fp)

        pipeline = PipelineModel(stages=[fp.transformer for fp in fitted_pipes])
        out_roles = dict()
        for fp in fitted_pipes:
            out_roles.update(fp.roles)

        return FittedPipe(sdf=current_sdf, transformer=pipeline, roles=out_roles)

    def _optimize_and_fit(self, train: SparkDataFrame, pipeline: SparkEstOrTrans) -> FittedPipe:
        graph = build_graph(pipeline)
        tr_layers = list(toposort.toposort(graph))

        logger.info(f"Number of layers in the current feature pipeline {self}: {len(tr_layers)}")

        fp_input_features = set(self.input_features)

        current_train: SparkDataFrame = train
        stages = []
        fp_output_cols: List[str] = []
        fp_output_roles: RolesDict = dict()
        for i, layer in enumerate(tr_layers):
            logger.debug(f"Calculating layer ({i + 1}/{len(tr_layers)}). The size of layer: {len(layer)}")
            cols_to_remove = []
            output_cols = []
            layer_model = Pipeline(stages=layer).fit(current_train)
            for j, tr in enumerate(layer):
                logger.debug(f"Processing output columns for transformer ({j + 1}/{len(layer)}): {tr}")
                tr = cast(SparkColumnsAndRoles, tr)
                if tr.getDoReplaceColumns():
                    # ChangeRoles, for instance, may return columns with the same name
                    # thus we don't want to remove these columns
                    self_out_cols = set(tr.getOutputCols())
                    cols_to_remove.extend([f for f in tr.getInputCols() if f not in self_out_cols])
                output_cols.extend(tr.getOutputCols())
                fp_output_roles.update(tr.getOutputRoles())
            fp_output_cols = [c for c in fp_output_cols if c not in cols_to_remove]
            fp_output_cols.extend(output_cols)

            # we cannot really remove input features thus we leave them in the dataframe
            # but they won't in features and roles
            cols_to_remove = set(c for c in cols_to_remove if c not in fp_input_features)

            cacher = Cacher(self._cacher_key)
            pipe = Pipeline(stages=[layer_model, DropColumnsTransformer(list(cols_to_remove)), cacher])
            stages.append(pipe.fit(current_train))
            current_train = cacher.dataset

        fp_output_roles = {f: fp_output_roles[f] for f in fp_output_cols}

        return FittedPipe(current_train, PipelineModel(stages=stages), roles=fp_output_roles)

    def release_cache(self):
        Cacher.release_cache_by_key(self._cacher_key)


class SparkTabularDataFeatures:
    """Helper class contains basic features transformations for tabular data.

    This method can de shared by all tabular feature pipelines,
    to simplify ``.create_automl`` definition.
    """

    def __init__(self, **kwargs: Any):
        """Set default parameters for tabular pipeline constructor.

        Args:
            **kwargs: Additional parameters.

        """
        self.multiclass_te_co = 3
        self.top_intersections = 5
        self.max_intersection_depth = 3
        self.subsample = 0.1  # 10000
        self.random_state = 42
        self.feats_imp = None
        self.ascending_by_cardinality = False

        self.max_bin_count = 10
        self.sparse_ohe = "auto"

        for k in kwargs:
            self.__dict__[k] = kwargs[k]

    def _get_input_features(self) -> Set[str]:
        raise NotImplementedError()

    def _cols_by_role(self, dataset: SparkDataset, role_name: str, **kwargs: Any) -> List[str]:
        cols = get_columns_by_role(dataset, role_name, **kwargs)
        filtered_cols = [col for col in cols if col in self._get_input_features()]
        return filtered_cols

    def get_cols_for_datetime(self, train: SparkDataset) -> Tuple[List[str], List[str]]:
        """Get datetime columns to calculate features.

        Args:
            train: Dataset with train data.

        Returns:
            2 list of features names - base dates and common dates.

        """
        base_dates = self._cols_by_role(train, "Datetime", base_date=True)
        datetimes = self._cols_by_role(train, "Datetime", base_date=False) + self._cols_by_role(
            train, "Datetime", base_date=True, base_feats=True
        )

        return base_dates, datetimes

    def get_datetime_diffs(self, train: SparkDataset) -> Optional[SparkBaseTransformer]:
        """Difference for all datetimes with base date.

        Args:
            train: Dataset with train data.

        Returns:
            Transformer or ``None`` if no required features.

        """
        base_dates, datetimes = self.get_cols_for_datetime(train)
        if len(datetimes) == 0 or len(base_dates) == 0:
            return None

        roles = {f: train.roles[f] for f in itertools.chain(base_dates, datetimes)}

        base_diff = SparkBaseDiffTransformer(input_roles=roles, base_names=base_dates, diff_names=datetimes)

        return base_diff

    def get_datetime_seasons(
        self, train: SparkDataset, outp_role: Optional[ColumnRole] = None
    ) -> Optional[SparkBaseTransformer]:
        """Get season params from dates.

        Args:
            train: Dataset with train data.
            outp_role: Role associated with output features.

        Returns:
            Transformer or ``None`` if no required features.

        """
        _, datetimes = self.get_cols_for_datetime(train)
        for col in copy(datetimes):
            if len(train.roles[col].seasonality) == 0 and train.roles[col].country is None:
                datetimes.remove(col)

        if len(datetimes) == 0:
            return

        if outp_role is None:
            outp_role = NumericRole(np.float32)

        roles = {f: train.roles[f] for f in datetimes}

        date_as_cat = SparkDateSeasonsTransformer(input_cols=datetimes, input_roles=roles, output_role=outp_role)

        return date_as_cat

    def get_numeric_data(
        self,
        train: SparkDataset,
        feats_to_select: Optional[List[str]] = None,
        prob: Optional[bool] = None,
    ) -> Optional[SparkBaseTransformer]:
        """Select numeric features.

        Args:
            train: Dataset with train data.
            feats_to_select: Features to handle. If ``None`` - default filter.
            prob: Probability flag.

        Returns:
            Transformer.

        """
        if feats_to_select is None:
            if prob is None:
                feats_to_select = self._cols_by_role(train, "Numeric")
            else:
                feats_to_select = self._cols_by_role(train, "Numeric", prob=prob)

        if len(feats_to_select) == 0:
            return None

        roles = {f: train.roles[f] for f in feats_to_select}

        num_processing = SparkChangeRolesTransformer(
            input_cols=feats_to_select, input_roles=roles, role=NumericRole(np.float32)
        )

        return num_processing

    def get_freq_encoding(
        self, train: SparkDataset, feats_to_select: Optional[List[str]] = None
    ) -> Optional[SparkBaseEstimator]:
        """Get frequency encoding part.

        Args:
            train: Dataset with train data.
            feats_to_select: Features to handle. If ``None`` - default filter.

        Returns:
            Transformer.

        """
        if feats_to_select is None:
            feats_to_select = self._cols_by_role(train, "Category", encoding_type="freq")

        if len(feats_to_select) == 0:
            return None

        roles = {f: train.roles[f] for f in feats_to_select}

        cat_processing = SparkFreqEncoderEstimator(input_cols=feats_to_select, input_roles=roles)

        return cat_processing

    def get_ordinal_encoding(
        self, train: SparkDataset, feats_to_select: Optional[List[str]] = None
    ) -> Optional[SparkBaseEstimator]:
        """Get order encoded part.

        Args:
            train: Dataset with train data.
            feats_to_select: Features to handle. If ``None`` - default filter.

        Returns:
            Transformer.

        """
        if feats_to_select is None:
            feats_to_select = self._cols_by_role(train, "Category", ordinal=True)

        if len(feats_to_select) == 0:
            return

        roles = {f: train.roles[f] for f in feats_to_select}

        ord = SparkOrdinalEncoderEstimator(
            input_cols=feats_to_select, input_roles=roles, subs=self.subsample, random_state=self.random_state
        )

        return ord

    def get_categorical_raw(
        self, train: SparkDataset, feats_to_select: Optional[List[str]] = None
    ) -> Optional[SparkBaseEstimator]:
        """Get label encoded categories data.

        Args:
            train: Dataset with train data.
            feats_to_select: Features to handle. If ``None`` - default filter.

        Returns:
            Transformer.

        """

        if feats_to_select is None:
            feats_to_select = []
            for i in ["auto", "oof", "int", "ohe"]:
                feats = self._cols_by_role(train, "Category", encoding_type=i)
                feats_to_select.extend(feats)

        if len(feats_to_select) == 0:
            return

        roles = {f: train.roles[f] for f in feats_to_select}

        cat_processing = SparkLabelEncoderEstimator(
            input_cols=feats_to_select, input_roles=roles, subs=self.subsample, random_state=self.random_state
        )
        return cat_processing

    def get_target_encoder(self, train: SparkDataset) -> Optional[type]:
        """Get target encoder func for dataset.

        Args:
            train: Dataset with train data.

        Returns:
            Class

        """
        target_encoder = None
        if train.folds is not None:
            if train.task.name in ["binary", "reg"]:
                target_encoder = SparkTargetEncoderEstimator
            else:
                result = train.data.select(F.max(train.target_column).alias("max")).first()
                n_classes = result["max"] + 1

                if n_classes <= self.multiclass_te_co:
                    target_encoder = SparkMulticlassTargetEncoderEstimator

        return target_encoder

    def get_binned_data(
        self, train: SparkDataset, feats_to_select: Optional[List[str]] = None
    ) -> Optional[SparkBaseEstimator]:
        """Get encoded quantiles of numeric features.

        Args:
            train: Dataset with train data.
            feats_to_select: features to hanlde. If ``None`` - default filter.

        Returns:
            Transformer.

        """
        if feats_to_select is None:
            feats_to_select = self._cols_by_role(train, "Numeric", discretization=True)

        if len(feats_to_select) == 0:
            return

        roles = {f: train.roles[f] for f in feats_to_select}

        binned_processing = SparkQuantileBinningEstimator(
            input_cols=feats_to_select, input_roles=roles, nbins=self.max_bin_count
        )

        return binned_processing

    def get_categorical_intersections(
        self, train: SparkDataset, feats_to_select: Optional[List[str]] = None
    ) -> Optional[SparkBaseEstimator]:
        """Get transformer that implements categorical intersections.

        Args:
            train: Dataset with train data.
            feats_to_select: features to handle. If ``None`` - default filter.

        Returns:
            Transformer.

        """

        if feats_to_select is None:
            categories = get_columns_by_role(train, "Category")
            feats_to_select = categories

            if len(categories) <= 1:
                return

            elif len(categories) > self.top_intersections:
                feats_to_select = self.get_top_categories(train, self.top_intersections)

        elif len(feats_to_select) <= 1:
            return

        roles = {f: train.roles[f] for f in feats_to_select}

        cat_processing = SparkCatIntersectionsEstimator(
            input_cols=feats_to_select, input_roles=roles, max_depth=self.max_intersection_depth
        )

        return cat_processing

    def get_uniques_cnt(self, train: SparkDataset, feats: List[str]) -> Series:
        """Get unique values cnt.

        Be aware that this function uses approx_count_distinct and thus cannot return precise results

        Args:
            train: Dataset with train data.
            feats: Features names.

        Returns:
            Series.

        """
        warn_if_not_cached(train.data)

        sdf = train.data.select(feats)

        # TODO SPARK-LAMA: Do we really need this sampling?
        # if self.subsample:
        #     sdf = sdf.sample(withReplacement=False, fraction=self.subsample, seed=self.random_state)

        sdf = sdf.select([F.approx_count_distinct(col).alias(col) for col in feats])
        result = sdf.collect()[0]

        uns = [result[col] for col in feats]
        return Series(uns, index=feats, dtype="int")

    def get_top_categories(self, train: SparkDataset, top_n: int = 5) -> List[str]:
        """Get top categories by importance.

        If feature importance is not defined,
        or feats has same importance - sort it by unique values counts.
        In second case init param ``ascending_by_cardinality``
        defines how - asc or desc.

        Args:
            train: Dataset with train data.
            top_n: Number of top categories.

        Returns:
            List.

        """
        if self.max_intersection_depth <= 1 or self.top_intersections <= 1:
            return []

        cats = get_columns_by_role(train, "Category")
        if len(cats) == 0:
            return []

        df = DataFrame({"importance": 0, "cardinality": 0}, index=cats)
        # importance if defined
        if self.feats_imp is not None:
            feats_imp = Series(self.feats_imp.get_features_score()).sort_values(ascending=False)
            df["importance"] = feats_imp[feats_imp.index.isin(cats)]
            df["importance"].fillna(-np.inf)

        # check for cardinality
        df["cardinality"] = self.get_uniques_cnt(train, cats)
        # sort
        df = df.sort_values(
            by=["importance", "cardinality"],
            ascending=[False, self.ascending_by_cardinality],
        )
        # get top n
        top = list(df.index[:top_n])

        return top


class SparkEmptyFeaturePipeline(SparkFeaturesPipeline):
    """
    This class creates pipeline with ``SparkNoOpTransformer``
    """

    def create_pipeline(self, train: SparkDataset) -> SparkEstOrTrans:
        """
        Returns ``SparkNoOpTransformer`` instance
        """
        return SparkNoOpTransformer()


class SparkNoOpTransformer(SparkBaseTransformer):
    """
    This transformer does nothing, it just returns the input dataframe unchanged.
    """

    def __init__(self):
        super().__init__(input_cols=[], output_cols=[], input_roles=dict(), output_roles=dict())

    def _transform(self, dataset):
        return dataset



import calendar
import datetime
import logging
from typing import Iterator, List, Tuple
import numpy as np

import pandas as pd

from pyspark.sql.pandas.functions import pandas_udf
from pyspark.sql.types import StructField

from sparklightautoml.utils import SparkDataFrame
from sparklightautoml.tasks.base import SparkMetric

logger = logging.getLogger(__name__)

MAX_DAY = {
    1: 31,
    2: 28,
    3: 31,
    4: 30,
    5: 31,
    6: 30,
    7: 31,
    8: 31,
    9: 30,
    10: 31,
    11: 30,
    12: 31,
}


def replace_year_in_date(date: datetime.datetime, year: int):
    if date.month == 2 and date.day == 29 and not calendar.isleap(year):
        date -= pd.Timedelta(1, "d")
    return date.replace(year=year)


def replace_month_in_date(date: datetime.datetime, month: int):
    if date.day > MAX_DAY[month]:
        date -= pd.Timedelta(date.day - MAX_DAY[month], "d")
        if month == 2 and date.day == 28 and calendar.isleap(date.year):
            date += pd.Timedelta(1, "d")
    return date.replace(month=month)


def replace_dayofweek_in_date(date: datetime.datetime, dayofweek: int):
    date += pd.Timedelta(dayofweek - date.weekday(), "d")
    return date


def calc_feats_permutation_imps(
    model, used_feats: List[str], data: SparkDataFrame, metric: SparkMetric, silent: bool = False
):
    n_used_feats = len(used_feats)
    if not silent:
        logger.info3("LightAutoML used {} feats".format(n_used_feats))

    preds = model.predict(data, add_reader_attrs=True)
    norm_score = metric(preds)

    feat_imp = []
    for it, feat in enumerate(used_feats):
        feat_imp.append(calc_one_feat_imp((it + 1, n_used_feats), feat, model, data, norm_score, metric, silent))
    feat_imp = pd.DataFrame(feat_imp, columns=["Feature", "Importance"])
    feat_imp = feat_imp.sort_values("Importance", ascending=False).reset_index(drop=True)
    return feat_imp


def calc_one_feat_imp(
    iters: Tuple[int, int],
    feat: str,
    model,
    data: SparkDataFrame,
    norm_score: float,
    metric: SparkMetric,
    silent: bool,
    seed: int = 42,
):

    field: StructField = data.schema[feat]

    @pandas_udf(returnType=field.dataType)
    def permutate(arrs: Iterator[pd.Series]) -> Iterator[pd.Series]:
        permutator = np.random.RandomState(seed=seed)
        # one may get list of arrs and concatenate them to perform permutation
        # in the whole partition
        for x in arrs:
            px = permutator.permutation(x)
            yield pd.Series(px)

    permutated_df = data.withColumn(feat, permutate(feat))

    preds = model.predict(permutated_df, add_reader_attrs=True)
    new_score = metric(preds)

    if not silent:
        logger.info3("{}/{} Calculated score for {}: {:.7f}".format(iters[0], iters[1], feat, norm_score - new_score))

    return feat, norm_score - new_score



import logging
from abc import ABC
from copy import copy, deepcopy
from typing import Dict, cast, Sequence, List, Set, Optional, Union

from pyspark.sql import Column
from pyspark.sql import functions as F

from sparklightautoml.spark_functions import array_to_vector
from lightautoml.dataset.base import RolesDict
from lightautoml.dataset.roles import ColumnRole
from lightautoml.dataset.utils import concatenate
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.mlwriters import CommonPickleMLReadable, CommonPickleMLWritable
from sparklightautoml.utils import log_exec_time, SparkDataFrame
from lightautoml.transformers.base import (
    LAMLTransformer,
    ColumnsSelector as LAMAColumnsSelector,
    ChangeRoles as LAMAChangeRoles,
)
from lightautoml.transformers.base import Roles

from pyspark.ml import Transformer, Estimator
from pyspark.ml.param.shared import HasInputCols, HasOutputCols, TypeConverters
from pyspark.ml.param.shared import Param, Params
from pyspark.ml.util import DefaultParamsWritable, DefaultParamsReadable


logger = logging.getLogger(__name__)


SparkEstOrTrans = Union[
    "SparkBaseEstimator", "SparkBaseTransformer", "SparkUnionTransformer", "SparkSequentialTransformer"
]


class HasInputRoles(Params):
    """
    Mixin for param inputCols: input column names.
    """

    inputRoles = Param(Params._dummy(), "inputRoles", "input roles (lama format)")

    def __init__(self):
        super().__init__()

    def getInputRoles(self):
        """
        Gets the value of inputCols or its default value.
        """
        return self.getOrDefault(self.inputRoles)


class HasOutputRoles(Params):
    """
    Mixin for param inputCols: input column names.
    """

    outputRoles = Param(Params._dummy(), "outputRoles", "output roles (lama format)")

    def __init__(self):
        super().__init__()

    def getOutputRoles(self):
        """
        Gets the value of inputCols or its default value.
        """
        return self.getOrDefault(self.outputRoles)


class SparkColumnsAndRoles(HasInputCols, HasOutputCols, HasInputRoles, HasOutputRoles):
    """
    Helper and base class for :class:`~sparklightautoml.transformers.base.SparkBaseTransformer` and
    :class:`~sparklightautoml.transformers.base.SparkBaseEstimator`.
    """

    doReplaceColumns = Param(Params._dummy(), "doReplaceColumns", "whatever it replaces columns or not")
    columnsToReplace = Param(Params._dummy(), "columnsToReplace", "which columns to replace")

    def getDoReplaceColumns(self) -> bool:
        return self.getOrDefault(self.doReplaceColumns)

    def getColumnsToReplace(self) -> List[str]:
        return self.getOrDefault(self.columnsToReplace)

    @staticmethod
    def make_dataset(
        transformer: "SparkColumnsAndRoles", base_dataset: SparkDataset, data: SparkDataFrame
    ) -> SparkDataset:
        new_roles = deepcopy(base_dataset.roles)
        new_roles.update(transformer.getOutputRoles())
        new_ds = base_dataset.empty()
        new_ds.set_data(data, base_dataset.features + transformer.getOutputCols(), new_roles)
        return new_ds


class SparkBaseEstimator(Estimator, SparkColumnsAndRoles, ABC):
    """
    Base class for estimators from `sparklightautoml.transformers`.
    """

    _fit_checks = ()
    _fname_prefix = ""

    def __init__(
        self,
        input_cols: Optional[List[str]] = None,
        input_roles: Optional[Dict[str, ColumnRole]] = None,
        do_replace_columns: bool = False,
        output_role: Optional[ColumnRole] = None,
    ):
        super().__init__()

        self._output_role = output_role

        assert all((f in input_roles) for f in input_cols), "All columns should have roles"

        self.set(self.inputCols, input_cols)
        self.set(self.outputCols, self._make_output_names(input_cols))
        self.set(self.inputRoles, input_roles)
        self.set(self.outputRoles, self._make_output_roles())
        self.set(self.doReplaceColumns, do_replace_columns)

    def _make_output_names(self, input_cols: List[str]) -> List[str]:
        return [f"{self._fname_prefix}__{feat}" for feat in input_cols]

    def _make_output_roles(self):
        assert self._output_role is not None
        new_roles = {}
        new_roles.update({feat: self._output_role for feat in self.getOutputCols()})
        return new_roles


class SparkBaseTransformer(Transformer, SparkColumnsAndRoles, ABC):
    """
    Base class for transformers from `sparklightautoml.transformers`.
    """

    _fname_prefix = ""

    def __init__(
        self,
        input_cols: List[str],
        output_cols: List[str],
        input_roles: RolesDict,
        output_roles: RolesDict,
        do_replace_columns: Union[bool, List[str]] = False,
    ):
        super().__init__()

        # assert len(input_cols) == len(output_cols)
        # assert len(input_roles) == len(output_roles)
        assert all((f in input_roles) for f in input_cols)
        assert all((f in output_roles) for f in output_cols)

        self.set(self.inputCols, input_cols)
        self.set(self.outputCols, output_cols)
        self.set(self.inputRoles, input_roles)
        self.set(self.outputRoles, output_roles)

        if isinstance(do_replace_columns, List):
            cols_to_replace = cast(List[str], do_replace_columns)
            assert (
                len(set(cols_to_replace).difference(set(self.getInputCols()))) == 0
            ), "All columns to replace, should be in input columns"
            self.set(self.doReplaceColumns, True)
            self.set(self.columnsToReplace, cols_to_replace)
        else:
            self.set(self.doReplaceColumns, do_replace_columns)
            self.set(self.columnsToReplace, self.getInputCols() if do_replace_columns else [])

    _transform_checks = ()

    def _make_output_df(self, input_df: SparkDataFrame, cols_to_add: List[Union[str, Column]]):
        return input_df.select("*", *cols_to_add)

        # if not self.getDoReplaceColumns():
        #     return input_df.select('*', *cols_to_add)
        #
        # replaced_columns = set(self.getColumnsToReplace())
        # cols_to_leave = [f for f in input_df.columns if f not in replaced_columns]
        # return input_df.select(*cols_to_leave, *cols_to_add)

    def transform(self, dataset, params=None):
        logger.info(f"In transformer {type(self)}. Columns: {sorted(dataset.columns)}")
        return super().transform(dataset, params)


class SparkUnionTransformer:
    """
    Entity that represents parallel layers (transformers) in preprocess pipeline.
    """

    def __init__(self, transformer_list: List[SparkEstOrTrans]):
        self._transformer_list = copy(transformer_list)

    @property
    def transformers(self) -> List[SparkEstOrTrans]:
        return self._transformer_list

    def _find_last_stage(self, stage):
        if isinstance(stage, SparkSequentialTransformer):
            stage = stage.transformers[-1]
            return stage
        return stage

    def get_output_cols(self) -> List[str]:
        """Get list of output columns from all stages

        Returns:
            List[str]: output columns
        """
        output_cols = []
        for stage in self._transformer_list:
            stage = self._find_last_stage(stage)
            output_cols.extend(stage.getOutputCols())
        return list(set(output_cols))

    def get_output_roles(self) -> RolesDict:
        """Get output roles from all stages

        Returns:
            RolesDict: output roles
        """
        roles = {}
        for stage in self._transformer_list:
            stage = self._find_last_stage(stage)
            roles.update(deepcopy(stage.getOutputRoles()))

        return roles


class SparkSequentialTransformer:
    """
    Entity that represents sequential of transformers in preprocess pipeline.
    """

    def __init__(self, transformer_list: List[SparkEstOrTrans]):
        self._transformer_list = copy(transformer_list)

    @property
    def transformers(self) -> List[SparkEstOrTrans]:
        return self._transformer_list


class ObsoleteSparkTransformer(LAMLTransformer):

    _features = []

    _can_unwind_parents: bool = True

    _input_features = []

    def get_output_names(self, input_cols: List[str]) -> List[str]:
        return [f"{self._fname_prefix}__{feat}" for feat in input_cols]

    def fit(self, dataset: SparkDataset, use_features: Optional[List[str]] = None) -> "ObsoleteSparkTransformer":

        logger.info(f"SparkTransformer of type: {type(self)}")

        if use_features:
            existing_feats = set(dataset.features)
            not_found_feats = [feat for feat in use_features if feat not in existing_feats]
            assert len(not_found_feats) == 0, f"Not found features {not_found_feats} among existing {existing_feats}"
            self._features = use_features
        else:
            self._features = dataset.features

        self._input_features = use_features if use_features else dataset.features

        for check_func in self._fit_checks:
            check_func(dataset, use_features)

        return self._fit(dataset)

    def _fit(self, dataset: SparkDataset) -> "ObsoleteSparkTransformer":
        return self

    def transform(self, dataset: SparkDataset) -> SparkDataset:
        for check_func in self._transform_checks:
            check_func(dataset)

        return self._transform(dataset)

    def _transform(self, dataset: SparkDataset) -> SparkDataset:
        return dataset

    def fit_transform(self, dataset: SparkDataset) -> SparkDataset:
        logger.info(f"fit_transform in {self._fname_prefix}: {type(self)}")

        self.fit(dataset)

        result = self.transform(dataset)

        return result

    @staticmethod
    def _get_updated_roles(dataset: SparkDataset, new_features: List[str], new_role: ColumnRole) -> RolesDict:
        new_roles = deepcopy(dataset.roles)
        new_roles.update({feat: new_role for feat in new_features})
        return new_roles


class ColumnsSelectorTransformer(
    Transformer, HasInputCols, HasOutputCols, DefaultParamsWritable, DefaultParamsReadable
):
    """
    Makes selection input columns from input dataframe.
    """

    optionalCols = Param(
        Params._dummy(), "optionalCols", "optional column names.", typeConverter=TypeConverters.toListString
    )

    def __init__(self, input_cols: Optional[List[str]] = [], optional_cols: Optional[List[str]] = []):
        super().__init__()
        optional_cols = optional_cols if optional_cols else []
        assert (
            len(set(input_cols).intersection(set(optional_cols))) == 0
        ), "Input columns and optional columns cannot intersect"

        self.set(self.inputCols, input_cols)
        self.set(self.optionalCols, optional_cols)
        self.set(self.outputCols, input_cols)

    def getOptionalCols(self) -> List[str]:
        return self.getOrDefault(self.optionalCols)

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:
        logger.info(f"In transformer {type(self)}. Columns: {sorted(dataset.columns)}")
        ds_cols = set(dataset.columns)
        present_opt_cols = [c for c in self.getOptionalCols() if c in ds_cols]
        return dataset.select(*self.getInputCols(), *present_opt_cols)


class ProbabilityColsTransformer(Transformer, DefaultParamsWritable, DefaultParamsReadable):
    """Converts probability columns values from ONNX model format to LGBMCBooster format"""

    probabilityCols = Param(
        Params._dummy(),
        "probabilityCols",
        "probability сolumn names in dataframe",
        typeConverter=TypeConverters.toListString,
    )

    numClasses = Param(Params._dummy(), "numClasses", "number of classes", typeConverter=TypeConverters.toInt)

    def __init__(self, probability_сols: List[str] = [], num_classes: int = 0):
        super().__init__()
        self.set(self.probabilityCols, probability_сols)
        self.set(self.numClasses, num_classes)

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:
        num_classes = self.getOrDefault(self.numClasses)
        probability_cols = self.getOrDefault(self.probabilityCols)
        other_cols = [c for c in dataset.columns if c not in probability_cols]
        probability_cols = [
            array_to_vector(F.array([F.col(c).getItem(i) for i in range(num_classes)])).alias(c)
            for c in probability_cols
        ]
        dataset = dataset.select(*other_cols, *probability_cols)
        return dataset


class PredictionColsTransformer(Transformer, DefaultParamsWritable, DefaultParamsReadable):
    """Converts prediction columns values from ONNX model format to LGBMCBooster format"""

    predictionCols = Param(
        Params._dummy(),
        "predictionCols",
        "probability сolumn names in dataframe",
        typeConverter=TypeConverters.toListString,
    )

    def __init__(self, prediction_сols: List[str] = []):
        super().__init__()
        self.set(self.predictionCols, prediction_сols)

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:
        prediction_cols = self.getOrDefault(self.predictionCols)
        other_cols = [c for c in dataset.columns if c not in prediction_cols]
        prediction_cols = [F.col(c).getItem(0).alias(c) for c in prediction_cols]
        dataset = dataset.select(*other_cols, *prediction_cols)
        return dataset


class DropColumnsTransformer(Transformer, DefaultParamsWritable, DefaultParamsReadable):
    """Transformer that drops columns from input dataframe."""

    colsToRemove = Param(
        Params._dummy(), "colsToRemove", "columns to be removed", typeConverter=TypeConverters.toListString
    )

    optionalColsToRemove = Param(
        Params._dummy(),
        "optionalColsToRemove",
        "optional columns to be removed (if they appear in the dataset)",
        typeConverter=TypeConverters.toListString,
    )

    def __init__(self, remove_cols: List[str] = [], optional_remove_cols: Optional[List[str]] = None):
        super().__init__()
        self.set(self.colsToRemove, remove_cols)
        self.set(self.optionalColsToRemove, optional_remove_cols if optional_remove_cols else [])

    def getColsToRemove(self) -> List[str]:
        return self.getOrDefault(self.colsToRemove)

    def getOptionalColsToRemove(self) -> List[str]:
        return self.getOrDefault(self.optionalColsToRemove)

    def _transform(self, dataset):
        ds_cols = set(dataset.columns)
        optional_to_remove = [c for c in self.getOptionalColsToRemove() if c in ds_cols]
        dataset = dataset.drop(*self.getColsToRemove(), *optional_to_remove)

        return dataset


class SparkChangeRolesTransformer(SparkBaseTransformer, CommonPickleMLWritable, CommonPickleMLReadable):
    """Transformer that change roles for input columns. Does not transform the input dataframe.

    # Note: this trasnformer cannot be applied directly to input columns of a feature pipeline
    """

    def __init__(self, input_cols: List[str], input_roles: RolesDict, role: ColumnRole):
        super().__init__(
            input_cols=input_cols,
            output_cols=input_cols,
            input_roles=input_roles,
            output_roles={f: deepcopy(role) for f in input_cols},
            do_replace_columns=True,
        )

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:
        return dataset



"""Linear models for tabular datasets."""

import logging
from copy import copy
from typing import Tuple, Optional, List
from typing import Union

import numpy as np
from pyspark.ml import Pipeline, Transformer, PipelineModel, Estimator
from pyspark.ml.classification import LogisticRegression, LogisticRegressionModel
from pyspark.ml.feature import VectorAssembler, OneHotEncoder
from pyspark.ml.regression import LinearRegression, LinearRegressionModel
from pyspark.sql import functions as F

from sparklightautoml.ml_algo.base import SparkTabularMLAlgo, SparkMLModel, AveragingTransformer
from sparklightautoml.validation.base import SparkBaseTrainValidIterator
from ..dataset.base import SparkDataset
from ..transformers.base import DropColumnsTransformer
from ..utils import DebugTransformer, SparkDataFrame
from lightautoml.utils.timer import TaskTimer

logger = logging.getLogger(__name__)

LinearEstimator = Union[LogisticRegression, LinearRegression]
LinearEstimatorModel = Union[LogisticRegressionModel, LinearRegressionModel]


class SparkLinearLBFGS(SparkTabularMLAlgo):
    """LBFGS L2 regression based on Spark MLlib.


    default_params:

        - tol: The tolerance for the stopping criteria.
        - maxIter: Maximum iterations of L-BFGS.
        - aggregationDepth: Param for suggested depth for treeAggregate.
        - elasticNetParam: Elastic net parameter.
        - regParam: Regularization parameter.
        - early_stopping: Maximum rounds without improving.

    freeze_defaults:

        - ``True`` :  params may be rewrited depending on dataset.
        - ``False``:  params may be changed only manually or with tuning.

    timer: :class:`~lightautoml.utils.timer.Timer` instance or ``None``.

    """

    _name: str = "LinearL2"

    _default_params = {
        "tol": 1e-6,
        "maxIter": 100,
        "aggregationDepth": 2,
        "elasticNetParam": 0.7,
        "regParam": [
            1e-5,
            5e-5,
            1e-4,
            5e-4,
            1e-3,
            5e-3,
            1e-2,
            5e-2,
            1e-1,
            5e-1,
            1,
            5,
            10,
            50,
            100,
            500,
            1000,
            5000,
            10000,
            50000,
            100000,
        ],
        "early_stopping": 2,
    }

    def __init__(
        self,
        cacher_key: str,
        default_params: Optional[dict] = None,
        freeze_defaults: bool = True,
        timer: Optional[TaskTimer] = None,
        optimization_search_space: Optional[dict] = {},
    ):
        super().__init__(cacher_key, default_params, freeze_defaults, timer, optimization_search_space)

        self._prediction_col = f"prediction_{self._name}"
        self.task = None
        self._timer = timer
        # self._ohe = None
        self._assembler = None

        self._raw_prediction_col_name = "raw_prediction"
        self._probability_col_name = "probability"
        self._prediction_col_name = "prediction"

    def _infer_params(
        self, train: SparkDataset, fold_prediction_column: str
    ) -> Tuple[List[Tuple[float, Estimator]], int]:
        logger.debug("Building pipeline in linear lGBFS")
        params = copy(self.params)

        if "regParam" in params:
            reg_params = params["regParam"]
            del params["regParam"]
        else:
            reg_params = [1.0]

        if "early_stopping" in params:
            es = params["early_stopping"]
            del params["early_stopping"]
        else:
            es = 100

        def build_pipeline(reg_param: int):
            instance_params = copy(params)
            instance_params["regParam"] = reg_param
            if self.task.name in ["binary", "multiclass"]:
                model = LogisticRegression(
                    featuresCol=self._assembler.getOutputCol(),
                    labelCol=train.target_column,
                    probabilityCol=fold_prediction_column,
                    rawPredictionCol=self._raw_prediction_col_name,
                    predictionCol=self._prediction_col_name,
                    **instance_params,
                )
            elif self.task.name == "reg":
                model = LinearRegression(
                    featuresCol=self._assembler.getOutputCol(),
                    labelCol=train.target_column,
                    predictionCol=fold_prediction_column,
                    **instance_params,
                )
                model = model.setSolver("l-bfgs")
            else:
                raise ValueError("Task not supported")

            return model

        estimators = [(rp, build_pipeline(rp)) for rp in reg_params]

        return estimators, es

    def fit_predict_single_fold(
        self, fold_prediction_column: str, full: SparkDataset, train: SparkDataset, valid: SparkDataset
    ) -> Tuple[SparkMLModel, SparkDataFrame, str]:
        """Train on train dataset and predict on holdout dataset.

        Args:
            fold_prediction_column: column name for predictions made for this fold
            full: Full dataset that include train and valid parts and a bool column that delimits records
            train: Train Dataset.
            valid: Validation Dataset.

        Returns:
            Target predictions for valid dataset.

        """
        logger.info(f"fit_predict single fold in LinearLBGFS. Num of features: {len(self.input_features)} ")

        if self.task is None:
            self.task = train.task

        train_sdf = train.data
        val_sdf = valid.data

        estimators, early_stopping = self._infer_params(train, fold_prediction_column)

        assert len(estimators) > 0

        es: int = 0
        best_score: float = -np.inf

        best_model: Optional[SparkMLModel] = None
        best_val_pred: Optional[SparkDataFrame] = None
        for rp, model in estimators:
            logger.debug(f"Fitting estimators with regParam {rp}")
            pipeline = Pipeline(stages=[self._assembler, model])
            ml_model = pipeline.fit(train_sdf)
            val_pred = ml_model.transform(val_sdf)
            preds_to_score = val_pred.select(
                F.col(fold_prediction_column).alias("prediction"), F.col(valid.target_column).alias("target")
            )
            current_score = self.score(preds_to_score)
            if current_score > best_score:
                best_score = current_score
                best_model = ml_model.stages[-1]
                best_val_pred = val_pred
                es = 0
            else:
                es += 1

            if es >= early_stopping:
                break

        logger.info("fit_predict single fold finished in LinearLBGFS")

        return best_model, best_val_pred, fold_prediction_column

    def predict_single_fold(self, dataset: SparkDataset, model: SparkMLModel) -> SparkDataFrame:
        """Implements prediction on single fold.

        Args:
            model: Model uses to predict.
            dataset: ``SparkDataset`` used for prediction.

        Returns:
            Predictions for input dataset.

        """
        pred = model.transform(dataset.data)
        return pred

    def _build_transformer(self) -> Transformer:
        avr = self._build_averaging_transformer()
        models = [
            el
            for m in self.models
            for el in [
                m,
                DropColumnsTransformer(
                    remove_cols=[],
                    optional_remove_cols=[
                        self._prediction_col_name,
                        self._probability_col_name,
                        self._raw_prediction_col_name,
                    ],
                ),
            ]
        ]
        averaging_model = PipelineModel(stages=[self._assembler] + models + [avr])
        return averaging_model

    def _build_averaging_transformer(self) -> Transformer:
        avr = AveragingTransformer(
            self.task.name,
            input_cols=self._models_prediction_columns,
            output_col=self.prediction_feature,
            remove_cols=[self._assembler.getOutputCol()] + self._models_prediction_columns,
            convert_to_array_first=not (self.task.name == "reg"),
            dim_num=self.n_classes,
        )
        return avr

    def fit_predict(self, train_valid_iterator: SparkBaseTrainValidIterator) -> SparkDataset:
        """Fit and then predict accordig the strategy that uses train_valid_iterator.

        If item uses more then one time it will
        predict mean value of predictions.
        If the element is not used in training then
        the prediction will be ``numpy.nan`` for this item

        Args:
            train_valid_iterator: Classic cv-iterator.

        Returns:
            Dataset with predicted values.

        """
        logger.info("Starting LinearLGBFS")
        self.timer.start()

        self.input_roles = train_valid_iterator.input_roles
        cat_feats = [feat for feat in self.input_features if self.input_roles[feat].name == "Category"]
        # self._ohe = OneHotEncoder(inputCols=cat_feats, outputCols=[f"{f}_{self._name}_ohe" for f in cat_feats])
        # self._ohe = self._ohe.fit(train_valid_iterator.train.data)

        non_cat_feats = [feat for feat in self.input_features if self.input_roles[feat].name != "Category"]
        self._assembler = VectorAssembler(
            inputCols=non_cat_feats + cat_feats,  # self._ohe.getOutputCols(),
            outputCol=f"{self._name}_vassembler_features",
        )

        result = super().fit_predict(train_valid_iterator)

        logger.info("LinearLGBFS is finished")

        return result



import logging
from copy import copy
from copy import deepcopy
from typing import Optional, Any, List, Dict, Tuple

import numpy as np
import pandas as pd
from pyspark.ml import Transformer
from pyspark.ml.param import Param, Params
from pyspark.sql import functions as F
from pyspark.sql.types import IntegerType, NumericType, FloatType, StringType

from lightautoml.dataset.base import array_attr_roles, valid_array_attributes
from lightautoml.dataset.roles import ColumnRole, DropRole, NumericRole, DatetimeRole, CategoryRole
from lightautoml.dataset.utils import roles_parser
from lightautoml.reader.base import Reader, UserDefinedRolesDict, RoleType, RolesDict
from lightautoml.reader.guess_roles import (
    calc_encoding_rules,
    rule_based_roles_guess,
    calc_category_rules,
    rule_based_cat_handler_guess,
)
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.mlwriters import CommonPickleMLReadable, CommonPickleMLWritable
from sparklightautoml.reader.guess_roles import get_numeric_roles_stat, get_category_roles_stat, get_null_scores
from sparklightautoml.utils import Cacher, SparkDataFrame
from lightautoml.tasks import Task

logger = logging.getLogger(__name__)

dtype2Stype = {
    "str": "string",
    "bool": "boolean",
    "int": "int",
    "int8": "int",
    "int16": "int",
    "int32": "int",
    "int64": "int",
    "int128": "bigint",
    "int256": "bigint",
    "integer": "int",
    "uint8": "int",
    "uint16": "int",
    "uint32": "int",
    "uint64": "int",
    "uint128": "bigint",
    "uint256": "bigint",
    "longlong": "long",
    "ulonglong": "long",
    "float16": "float",
    "float": "float",
    "float32": "float",
    "float64": "double",
    "float128": "double",
}

stype2dtype = {
    "string": "str",
    "boolean": "bool",
    "bool": "bool",
    "int": "int",
    "bigint": "longlong",
    "long": "long",
    "float": "float",
    "double": "float64",
}


class SparkReaderHelper:
    """Helper class that provide some methods for :class:`~sparklightautoml.reader.base.SparkToSparkReader` and
    :class:`~sparklightautoml.reader.base.SparkToSparkReaderTransformer`.
    """

    @staticmethod
    def _create_unique_ids(train_data: SparkDataFrame, cacher_key: Optional[str] = None) -> SparkDataFrame:
        logger.debug("SparkReaderHelper._create_unique_ids() is started")

        if SparkDataset.ID_COLUMN not in train_data.columns:
            train_data = train_data.select("*", F.monotonically_increasing_id().alias(SparkDataset.ID_COLUMN))

        if cacher_key is not None:
            cacher = Cacher(key=cacher_key)
            cacher.fit(train_data)
            train_data = cacher.dataset

        logger.debug("SparkReaderHelper._create_unique_ids() is finished")

        return train_data

    @staticmethod
    def _convert_column(feat: str, role: ColumnRole):
        if isinstance(role, DatetimeRole):
            result_column = F.to_timestamp(feat, role.format).alias(feat)
        elif isinstance(role, NumericRole):
            typ = dtype2Stype[role.dtype.__name__]
            result_column = F.when(F.isnull(feat), float("nan")).otherwise(F.col(feat).astype(typ)).alias(feat)
        else:
            result_column = F.col(feat)

        return result_column

    @staticmethod
    def _process_target_column(
        task_name: str, class_mapping: Dict, sdf: SparkDataFrame, target_col: str
    ) -> SparkDataFrame:
        if class_mapping is not None:
            sdf = sdf.replace(class_mapping, subset=[target_col])

        to_type = FloatType() if task_name == "reg" else IntegerType()

        cols = copy(sdf.columns)
        cols.remove(target_col)
        sdf = sdf.select(*cols, F.col(target_col).astype(to_type).alias(target_col))

        return sdf


class SparkToSparkReader(Reader, SparkReaderHelper):
    """
    Reader to convert :class:`~pandas.DataFrame` to AutoML's :class:`~lightautoml.dataset.np_pd_dataset.PandasDataset`.
    Stages:

        - Drop obviously useless features.
        - Convert roles dict from user format to automl format.
        - Simple role guess for features without input role.
        - Create cv folds.
        - Create initial PandasDataset.
        - Optional: advanced guessing of role and handling types.

    """

    DEFAULT_READER_FOLD_COL = "reader_fold_num"

    def __init__(
        self,
        task: Task,
        samples: Optional[int] = 100000,
        max_nan_rate: float = 0.999,
        max_constant_rate: float = 0.999,
        cv: int = 5,
        random_state: int = 42,
        roles_params: Optional[dict] = None,
        n_jobs: int = 4,
        # params for advanced roles guess
        advanced_roles: bool = True,
        numeric_unique_rate: float = 0.999,
        max_to_3rd_rate: float = 1.1,
        binning_enc_rate: float = 2,
        raw_decr_rate: float = 1.1,
        max_score_rate: float = 0.2,
        abs_score_val: float = 0.04,
        drop_score_co: float = 0.01,
        cacher_key: str = "default_cacher",
        **kwargs: Any,
    ):
        """

        Args:
            task: Task object.
            samples: Number of elements used when checking role type.
            max_nan_rate: Maximum nan-rate.
            max_constant_rate: Maximum constant rate.
            cv: CV Folds.
            random_state: Random seed.
            roles_params: dict of params of features roles. \
                Ex. {'numeric': {'dtype': np.float32}, 'datetime': {'date_format': '%Y-%m-%d'}}
                It's optional and commonly comes from config
            n_jobs: Int number of processes.
            advanced_roles: Param of roles guess (experimental, do not change).
            numeric_unqiue_rate: Param of roles guess (experimental, do not change).
            max_to_3rd_rate: Param of roles guess (experimental, do not change).
            binning_enc_rate: Param of roles guess (experimental, do not change).
            raw_decr_rate: Param of roles guess (experimental, do not change).
            max_score_rate: Param of roles guess (experimental, do not change).
            abs_score_val: Param of roles guess (experimental, do not change).
            drop_score_co: Param of roles guess (experimental, do not change).
            **kwargs: For now not used.

        """
        super().__init__(task)
        self.samples = samples
        self.max_nan_rate = max_nan_rate
        self.max_constant_rate = max_constant_rate
        self.cv = cv
        self.random_state = random_state
        self.n_jobs = n_jobs

        self.roles_params = roles_params
        self.target = None
        self.target_col = None
        if roles_params is None:
            self.roles_params = {}

        self.advanced_roles = advanced_roles
        self.advanced_roles_params = {
            "numeric_unique_rate": numeric_unique_rate,
            "max_to_3rd_rate": max_to_3rd_rate,
            "binning_enc_rate": binning_enc_rate,
            "raw_decr_rate": raw_decr_rate,
            "max_score_rate": max_score_rate,
            "abs_score_val": abs_score_val,
            "drop_score_co": drop_score_co,
        }

        self._cacher_key = cacher_key

        self.params = kwargs

    def fit_read(
        self, train_data: SparkDataFrame, features_names: Any = None, roles: UserDefinedRolesDict = None, **kwargs: Any
    ) -> SparkDataset:
        """Get dataset with initial feature selection.

        Args:
            train_data: Input data.
            features_names: Ignored. Just to keep signature.
            roles: Dict of features roles in format
              ``{RoleX: ['feat0', 'feat1', ...], RoleY: 'TARGET', ....}``.
            **kwargs: Can be used for target/group/weights.

        Returns:
            Dataset with selected features.

        """
        logger.info("Reader starting fit_read")
        logger.info(f"\x1b[1mTrain data columns: {train_data.columns}\x1b[0m\n")

        train_data = self._create_unique_ids(train_data, cacher_key=self._cacher_key)

        if roles is None:
            roles = {}

        # transform roles from user format {RoleX: ['feat0', 'feat1', ...], RoleY: 'TARGET', ....}
        # to automl format {'feat0': RoleX, 'feat1': RoleX, 'TARGET': RoleY, ...}
        parsed_roles = roles_parser(roles)
        # transform str role definition to automl ColumnRole
        attrs_dict = dict(zip(array_attr_roles, valid_array_attributes))

        for feat in parsed_roles:
            r = parsed_roles[feat]
            if type(r) == str:
                # get default role params if defined
                r = self._get_default_role_from_str(r)

            # check if column is defined like target/group/weight etc ...
            if r.name in attrs_dict:
                # defined in kwargs is rewrited.. TODO: Maybe raise warning if rewrited?
                # TODO: Think, what if multilabel or multitask? Multiple column target ..
                # TODO: Maybe for multilabel/multitask make target only avaliable in kwargs??
                self._used_array_attrs[attrs_dict[r.name]] = feat
                kwargs[attrs_dict[r.name]] = feat
                r = DropRole()

            # add new role
            parsed_roles[feat] = r

        assert "target" in kwargs, "Target should be defined"
        assert isinstance(kwargs["target"], str), f"Target must be a column name, but it is {type(kwargs['target'])}"
        assert kwargs["target"] in train_data.columns, f"Target must be a part of dataframe. Target: {kwargs['target']}"
        self.target_col = kwargs["target"]

        train_data = self._create_target(train_data, target_col=self.target_col)

        total_number = train_data.count()
        # if self.samples is not None:
        #     if self.samples > total_number:
        #         fraction = 1.0
        #     else:
        #         fraction = self.samples/total_number
        #     subsample = train_data.sample(fraction=fraction, seed=self.random_state).cache()
        # else:
        subsample = train_data

        logger.debug("SparkToSparkReader infer roles is started")
        # infer roles
        feats_to_guess: List[str] = []
        inferred_feats: Dict[str, ColumnRole] = dict()
        for feat in subsample.columns:
            if feat in [SparkDataset.ID_COLUMN, self.target_col]:
                continue

            assert isinstance(
                feat, str
            ), "Feature names must be string," " find feature name: {}, with type: {}".format(feat, type(feat))
            if feat in parsed_roles:
                r = parsed_roles[feat]
                # handle datetimes
                if r.name == "Datetime":
                    # try if it's ok to infer date with given params
                    result = subsample.select(
                        F.sum(F.to_timestamp(feat, format=r.format).isNotNull().astype(IntegerType())).alias(
                            f"{feat}_dt"
                        ),
                        F.count("*").alias("count"),
                    ).first()

                    if result[f"{feat}_dt"] != result["count"]:
                        raise ValueError("Looks like given datetime parsing params are not correctly defined")

                # replace default category dtype for numeric roles dtype if cat col dtype is numeric
                if r.name == "Category":
                    # default category role
                    cat_role = self._get_default_role_from_str("category")
                    # check if role with dtypes was exactly defined
                    flg_default_params = feat in roles["category"] if "category" in roles else False

                    inferred_dtype = next(dtyp for fname, dtyp in subsample.dtypes if fname == feat)
                    inferred_dtype = np.dtype(stype2dtype[inferred_dtype])

                    if (
                        flg_default_params
                        and not np.issubdtype(cat_role.dtype, np.number)
                        and np.issubdtype(inferred_dtype, np.number)
                    ):
                        r.dtype = self._get_default_role_from_str("numeric").dtype

                inferred_feats[feat] = r
            else:
                feats_to_guess.append(feat)

        logger.debug("SparkToSparkReader infer roles is finished")

        ok_features = self._ok_features(train_data, feats_to_guess)
        guessed_feats = self._guess_role(subsample, ok_features)
        inferred_feats.update(guessed_feats)

        # # set back
        for feat, r in inferred_feats.items():
            if r.name != "Drop":
                self._roles[feat] = r
                self._used_features.append(feat)
            else:
                self._dropped_features.append(feat)

        assert len(self.used_features) > 0, "All features are excluded for some reasons"

        # create folds
        train_data, folds_col = self._create_folds(train_data, kwargs)

        kwargs["folds"] = folds_col
        kwargs["target"] = self.target_col

        ff = [
            F.when(F.isnull(f), float("nan")).otherwise(F.col(f).astype(FloatType())).alias(f)
            if isinstance(self.roles[f], NumericRole)
            else f
            for f in self.used_features
        ]

        train_data = train_data.select(SparkDataset.ID_COLUMN, self.target_col, folds_col, *ff)

        dataset = SparkDataset(train_data, self.roles, task=self.task, **kwargs)

        if self.advanced_roles:
            new_roles = self.advanced_roles_guess(dataset, manual_roles=parsed_roles)

            droplist = [x for x in new_roles if new_roles[x].name == "Drop" and not self._roles[x].force_input]
            self.upd_used_features(remove=droplist)
            self._roles = {x: new_roles[x] for x in new_roles if x not in droplist}

            dataset = SparkDataset(
                train_data.select(SparkDataset.ID_COLUMN, *self.used_features), self.roles, task=self.task, **kwargs
            )

        logger.info("Reader finished fit_read")

        return dataset

    def read(self, data: SparkDataFrame, features_names: Any = None, add_array_attrs: bool = False) -> SparkDataset:
        """Read dataset with fitted metadata.

        Args:
            data: Data.
            features_names: Not used.
            add_array_attrs: Additional attributes, like
              target/group/weights/folds.

        Returns:
            Dataset with new columns.

        """

        kwargs = {}
        if add_array_attrs:
            for array_attr in self.used_array_attrs:
                col_name = self.used_array_attrs[array_attr]
                if col_name not in data.columns:
                    continue
                kwargs[array_attr] = col_name

        transformer = self.make_transformer(add_array_attrs)
        data = transformer.transform(data)

        dataset = SparkDataset(data, roles=self.roles, task=self.task, **kwargs)

        return dataset

    def make_transformer(self, add_array_attrs: bool = False):
        roles = {f: self.roles[f] for f in self.used_features}
        transformer = SparkToSparkReaderTransformer(
            self.task.name, self.class_mapping, copy(self.used_array_attrs), roles, add_array_attrs
        )
        return transformer

    def _create_folds(self, sdf: SparkDataFrame, kwargs: dict) -> Tuple[SparkDataFrame, str]:
        """
        Checks or adds folds column into the data frame
        Args:
            sdf: data frame to check or generate folds column for
            kwargs: to look folds column name

        Returns:
            The dataframe with the folds column, folds column name
        """
        if "folds" in kwargs:
            folds_col = kwargs["folds"]

            assert isinstance(folds_col, str), (
                f"If kwargs contains 'folds' it should be of type str and contain folds column."
                f"But kwargs['folds'] has type {type(folds_col)} and contains {folds_col}"
            )

            assert folds_col in sdf.columns, (
                f"Folds column ({folds_col}) should be presented in the train dataframe,"
                f"but it is not possible to find the column among {sdf.columns}"
            )

            cols_dtypes = dict(sdf.dtypes)
            assert (
                cols_dtypes[folds_col] == "int"
            ), f"Folds column should be of integer type, but it is {cols_dtypes[folds_col]}"

            return sdf, folds_col

        h = 1.0 / self.cv
        folds_col = self.DEFAULT_READER_FOLD_COL
        sdf_with_folds = sdf.select("*", F.floor(F.rand(self.random_state) / h).alias(folds_col))
        return sdf_with_folds, folds_col

    def _create_target(self, sdf: SparkDataFrame, target_col: str = "target") -> SparkDataFrame:
        """Validate target column and create class mapping if needed

        Args:
            target: Column with target values.

        Returns:
            Transformed target.

        """
        logger.debug("SparkToSparkReader._create_target() is started")

        self.class_mapping = None

        nan_count = sdf.where(F.isnan(target_col)).count()
        assert nan_count == 0, "Nan in target detected"

        if self.task.name != "reg":
            uniques = sdf.select(target_col).distinct().collect()
            uniques = [r[target_col] for r in uniques]
            self._n_classes = len(uniques)

            if isinstance(sdf.schema[target_col].dataType, NumericType):
                uniques = sorted(uniques)
                self.class_mapping = {x: i for i, x in enumerate(uniques)}
                srtd = np.ndarray(uniques)
            elif isinstance(sdf.schema[target_col].dataType, StringType):
                self.class_mapping = {x: f"{i}" for i, x in enumerate(uniques)}
                srtd = None
            else:
                raise ValueError(
                    f"Unsupported type of target column {sdf.schema[target_col]}. "
                    f"Only numeric and string are supported."
                )

            if srtd and (np.arange(srtd.shape[0]) == srtd).all():

                assert srtd.shape[0] > 1, "Less than 2 unique values in target"
                if self.task.name == "binary":
                    assert srtd.shape[0] == 2, "Binary task and more than 2 values in target"
                return sdf

        sdf_with_proc_target = self._process_target_column(self.task.name, self.class_mapping, sdf, target_col)

        logger.debug("SparkToSparkReader._create_target() is finished")

        return sdf_with_proc_target

    def _get_default_role_from_str(self, name) -> RoleType:
        """Get default role for string name according to automl's defaults and user settings.

        Args:
            name: name of role to get.

        Returns:
            role object.

        """
        name = name.lower()
        try:
            role_params = self.roles_params[name]
        except KeyError:
            role_params = {}

        return ColumnRole.from_string(name, **role_params)

    def _guess_role(self, data: SparkDataFrame, features: List[Tuple[str, bool]]) -> Dict[str, RoleType]:
        """Try to infer role, simple way.

        If convertable to float -> number.
        Else if convertable to datetime -> datetime.
        Else category.

        Args:
            feature: Column from dataset.

        Returns:
            Feature role.

        """
        logger.debug("SparkToSparkReader._guess_role() is started")

        guessed_cols = dict()
        cols_to_check = []
        check_columns = []

        feat2dtype = dict(data.dtypes)

        for feature, ok in features:
            if not ok:
                guessed_cols[feature] = DropRole()
                continue
            inferred_dtype = feat2dtype[feature]
            # numpy doesn't understand 'string' but 'str' is ok
            inferred_dtype = np.dtype(stype2dtype[inferred_dtype])

            # testing if it can be numeric or not
            num_dtype = self._get_default_role_from_str("numeric").dtype
            date_format = self._get_default_role_from_str("datetime").format
            # TODO: can it be really converted?
            if np.issubdtype(inferred_dtype, bool):
                guessed_cols[feature] = CategoryRole(dtype=np.bool8)
                continue
            elif np.issubdtype(inferred_dtype, np.number):
                guessed_cols[feature] = NumericRole(num_dtype)
                continue

            fcol = F.col(feature)

            can_cast_to_numeric = (
                F.when(F.isnull(fcol), True)
                .otherwise(fcol.cast(dtype2Stype[num_dtype.__name__]).isNotNull())
                .astype(IntegerType())
            )

            # TODO: utc handling here?
            can_cast_to_datetime = F.to_timestamp(feature, format=date_format).isNotNull().astype(IntegerType())

            cols_to_check.append((feature, num_dtype, date_format))
            check_columns.extend(
                [
                    F.sum(can_cast_to_numeric).alias(f"{feature}_num"),
                    F.sum(can_cast_to_datetime).alias(f"{feature}_dt"),
                ]
            )

        result = data.select(*check_columns, F.count("*").alias("count")).first()

        for feature, num_dtype, date_format in cols_to_check:
            if result[f"{feature}_num"] == result["count"]:
                guessed_cols[feature] = NumericRole(num_dtype)
            elif result[f"{feature}_dt"] == result["count"]:
                guessed_cols[feature] = DatetimeRole(np.datetime64, date_format=date_format)
            else:
                guessed_cols[feature] = CategoryRole(object)

        logger.debug("SparkToSparkReader._guess_role() is finished")

        return guessed_cols

    def _ok_features(self, train_data: SparkDataFrame, features: List[str]) -> List[Tuple[str, bool]]:
        """Check if column is filled well to be a feature.

        Args:
            feature: Column from dataset.

        Returns:
            ``True`` if nan ratio and freqency are not high.

        """
        logger.debug("SparkToSparkReader._ok_features() is started")

        row = train_data.select(
            F.count("*").alias("count"),
            *[
                F.mean((F.isnull(feature) | F.isnan(feature)).astype(IntegerType())).alias(f"{feature}_nan_rate")
                for feature in features
                if isinstance(train_data.schema[feature].dataType, NumericType)
            ],
            *[
                F.mean((F.isnull(feature)).astype(IntegerType())).alias(f"{feature}_nan_rate")
                for feature in features
                if not isinstance(train_data.schema[feature].dataType, NumericType)
            ],
        ).first()

        estimated_features = []
        for feat in features:
            if row[f"{feat}_nan_rate"] >= self.max_nan_rate:
                estimated_features.append((feat, False))
                continue

            # TODO: this part may be optimized using sampling
            crow = (
                train_data.groupby(feat)
                .agg(F.count("*").alias("count"))
                .select((F.max("count")).alias("count"))
                .first()
            )
            if crow["count"] / row["count"] >= self.max_constant_rate:
                estimated_features.append((feat, False))
                continue

            estimated_features.append((feat, True))

        logger.debug("SparkToSparkReader._ok_features() is finished")
        return estimated_features

    def advanced_roles_guess(self, dataset: SparkDataset, manual_roles: Optional[RolesDict] = None) -> RolesDict:
        """Advanced roles guess over user's definition and reader's simple guessing.

        Strategy - compute feature's NormalizedGini
        for different encoding ways and calc stats over results.
        Role is inferred by comparing performance stats with manual rules.
        Rule params are params of roles guess in init.
        Defaults are ok in general case.

        Args:
            dataset: Input PandasDataset.
            manual_roles: Dict of user defined roles.

        Returns:
            Dict.

        """
        logger.info("AdvGuessRoles: Calculating advanced guess roles")

        if manual_roles is None:
            manual_roles = {}
        top_scores = []
        new_roles_dict = dataset.roles
        advanced_roles_params = deepcopy(self.advanced_roles_params)
        drop_co = advanced_roles_params.pop("drop_score_co")
        # guess roles nor numerics

        stat = get_numeric_roles_stat(
            dataset,
            manual_roles=manual_roles,
            random_state=self.random_state,
            subsample=self.samples,
        )

        logger.info("AdvGuessRoles: Numeric roles stats were calculated")

        if len(stat) > 0:
            # upd stat with rules

            stat = calc_encoding_rules(stat, **advanced_roles_params)
            new_roles_dict = {**new_roles_dict, **rule_based_roles_guess(stat)}
            top_scores.append(stat["max_score"])

        # # # guess categories handling type
        stat = get_category_roles_stat(dataset, random_state=self.random_state, subsample=self.samples)
        if len(stat) > 0:
            # upd stat with rules
            # TODO: add sample params

            stat = calc_category_rules(stat)
            new_roles_dict = {**new_roles_dict, **rule_based_cat_handler_guess(stat)}
            top_scores.append(stat["max_score"])

        logger.info("AdvGuessRoles: Category roles stats were calculated")

        #
        # # get top scores of feature
        if len(top_scores) > 0:
            top_scores = pd.concat(top_scores, axis=0)
            # TODO: Add sample params

            null_scores = get_null_scores(
                dataset, list(top_scores.index), random_state=self.random_state, subsample=self.samples
            )

            logger.info("AdvGuessRoles: Null scores stats were calculated")
            top_scores = pd.concat([null_scores, top_scores], axis=1).max(axis=1)
            rejected = list(top_scores[top_scores < drop_co].index)
            logger.info("Feats was rejected during automatic roles guess: {0}".format(rejected))
            new_roles_dict = {**new_roles_dict, **{x: DropRole() for x in rejected}}

        return new_roles_dict


class SparkToSparkReaderTransformer(Transformer, SparkReaderHelper, CommonPickleMLWritable, CommonPickleMLReadable):
    """
    Transformer of SparkToSparkReader. Allows to reuse SparkToSparkReader pipeline as a spark transformer.
    """

    usedArrayAttrs = Param(Params._dummy(), "usedArrayAttrs", "usedArrayAttrs")
    addArrayAttrs = Param(Params._dummy(), "addArrayAttrs", "addArrayAttrs")
    roles = Param(Params._dummy(), "roles", "roles")
    taskName = Param(Params._dummy(), "taskName", "task name")
    classMapping = Param(Params._dummy(), "classMapping", "class mapping")

    def __init__(
        self,
        task_name: str,
        class_mapping: Optional[Dict],
        used_array_attrs: Dict[str, str],
        roles: Dict[str, ColumnRole],
        add_array_attrs: bool = False,
    ):
        super().__init__()
        self.set(self.taskName, task_name)
        self.set(self.classMapping, class_mapping)
        self.set(self.usedArrayAttrs, used_array_attrs)
        self.set(self.roles, roles)
        self.set(self.addArrayAttrs, add_array_attrs)

    def getTaskName(self) -> str:
        return self.getOrDefault(self.taskName)

    def getClassMapping(self) -> Optional[Dict]:
        return self.getOrDefault(self.classMapping)

    def getRoles(self) -> Dict[str, ColumnRole]:
        return self.getOrDefault(self.roles)

    def getUsedArrayAttrs(self) -> Dict[str, str]:
        return self.getOrDefault(self.usedArrayAttrs)

    def getAddArrayAttrs(self) -> bool:
        return self.getOrDefault(self.addArrayAttrs)

    def setAddArrayAttrs(self, value: bool):
        return self.set(self.addArrayAttrs, value)

    def _transform(self, data: SparkDataFrame) -> SparkDataFrame:
        service_columns = []

        used_array_attrs = self.getUsedArrayAttrs()
        roles = self.getRoles()

        if self.getAddArrayAttrs():
            for array_attr in used_array_attrs:
                col_name = used_array_attrs[array_attr]

                if col_name not in data.columns:
                    continue

                if array_attr == "target":
                    data = self._process_target_column(self.getTaskName(), self.getClassMapping(), data, col_name)

                service_columns.append(col_name)

        data = self._create_unique_ids(data)

        data = data.select(
            SparkDataset.ID_COLUMN,
            *service_columns,
            *[self._convert_column(feat, role) for feat, role in roles.items()],
        )

        return data
